from __future__ import annotations

import logging
import weakref

import numpy
from silx.gui import qt

from nxtomo.nxobject.nxdetector import ImageKey, FOV
from nxtomo.utils.transformation import build_matrix

from tomwer.core.process.edit.nxtomoeditor import NXtomoEditorTask, NXtomoEditorKeys
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.utils.scandescription import ScanNameLabelAndShape
from tomwer.gui.utils.unitsystem import MetricEntry
from tomwer.gui.edit.nxtomowarmer import NXtomoProxyWarmer

_logger = logging.getLogger(__name__)


class NXtomoEditorDialog(qt.QDialog):
    """
    Dialog embedding instances of NXtomoEditor and NXtomoProxyWarmer
    """

    def __init__(self, parent=None, hide_lockers=True) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self.mainWidget = NXtomoEditor(parent=self, hide_lockers=hide_lockers)
        self.layout().addWidget(self.mainWidget)
        self._warmer = NXtomoProxyWarmer(parent=self)
        self.layout().addWidget(self._warmer)

        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("validate")
        self.layout().addWidget(self._buttons)

    # expose API
    def setScan(self, scan: NXtomoScan) -> None:
        self.mainWidget.setScan(scan)
        self._warmer.setScan(scan)

    def hasLockField(self) -> bool:
        return self.mainWidget.hasLockField()

    def getConfiguration(self) -> dict:
        return self.mainWidget.getConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self.mainWidget.setConfiguration(config)

    def getConfigurationForTask(self) -> dict:
        return self.mainWidget.getConfigurationForTask()


class NXtomoEditor(qt.QWidget):
    """
    Widget to edit a set of field from a NXtomo.
    The preliminary goal is to let the user define pixel / voxel position and x and z positions
    in order to simplify stitching down the line

    As energy and field of view was also often requested this part is also editable.

    Each field contains a widget to define it values and a LockButton. The LockButton can be used to automate the
    processing.

    When the scan to edit is set each field widget will be updated **at the condition** the field is not locked.
    Else existing value will be kept.
    """

    sigEditingFinished = qt.Signal()
    """emit when edition is finished"""

    def __init__(self, parent=None, hide_lockers=True):
        super().__init__(parent)
        self._editableWidgets = []
        self._lockerPBs = []
        # list of all lockers
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._scanInfoQLE = ScanNameLabelAndShape(parent=self)
        self.layout().addWidget(self._scanInfoQLE)

        # nxtomo tree
        self._tree = qt.QTreeWidget(self)
        if hide_lockers:
            self._tree.setColumnCount(2)
            self._tree.setHeaderLabels(("entry", "value"))
        else:
            self._tree.setColumnCount(3)
            self._tree.setHeaderLabels(("entry", "value", "lockers"))
            self._tree.header().setStretchLastSection(False)
            self._tree.setColumnWidth(2, 20)
            self._tree.header().setSectionResizeMode(1, qt.QHeaderView.Stretch)
        self.layout().addWidget(self._tree)

        # 1: instrument
        self._instrumentQTWI = qt.QTreeWidgetItem(self._tree)
        self._instrumentQTWI.setText(0, "instrument")
        # handle energy
        self._beamQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._beamQTWI.setText(0, "beam")
        self._energyQTWI = qt.QTreeWidgetItem(self._beamQTWI)
        self._energyQTWI.setText(0, "energy (keV)")
        self._energyEntry = EnergyEntry("", self)
        self._energyEntry.setPlaceholderText("energy in kev")
        self._tree.setItemWidget(self._energyQTWI, 1, self._energyEntry)
        self._editableWidgets.append(self._energyEntry)
        self._energyLockerLB = PadlockButton(self)
        self._energyLockerLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._energyLockerLB)
        self._tree.setItemWidget(self._energyQTWI, 2, self._energyLockerLB)

        # 1.1 detector
        self._detectorQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._detectorQTWI.setText(0, "detector")
        ## pixel size
        self._xPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._xPixelSizeQTWI.setText(0, "x pixel size")
        self._xPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._xPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(self._xPixelSizeQTWI, 1, self._xPixelSizeMetricEntry)
        self._editableWidgets.append(self._xPixelSizeMetricEntry)
        self._xPixelSizeLB = PadlockButton(self)
        self._xPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xPixelSizeLB)
        self._tree.setItemWidget(self._xPixelSizeQTWI, 2, self._xPixelSizeLB)

        self._yPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._yPixelSizeQTWI.setText(0, "y pixel size")
        self._yPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._yPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(self._yPixelSizeQTWI, 1, self._yPixelSizeMetricEntry)
        self._editableWidgets.append(self._yPixelSizeMetricEntry)
        self._yPixelSizeLB = PadlockButton(self)
        self._yPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._yPixelSizeLB)
        self._tree.setItemWidget(self._yPixelSizeQTWI, 2, self._yPixelSizeLB)

        ## distance
        self._sampleDetectorDistanceQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._sampleDetectorDistanceQTWI.setText(0, "distance")
        self._distanceMetricEntry = MetricEntry("", parent=self)
        self._distanceMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._sampleDetectorDistanceQTWI, 1, self._distanceMetricEntry
        )
        self._editableWidgets.append(self._distanceMetricEntry)
        self._distanceLB = PadlockButton(self)
        self._distanceLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._distanceLB)
        self._tree.setItemWidget(self._sampleDetectorDistanceQTWI, 2, self._distanceLB)

        ## field of view
        self._fieldOfViewQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._fieldOfViewQTWI.setText(0, "field of view")
        self._fieldOfViewCB = qt.QComboBox(self)
        for value in FOV.values():
            self._fieldOfViewCB.addItem(value)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 1, self._fieldOfViewCB)
        self._editableWidgets.append(self._fieldOfViewCB)
        self._fieldOfViewLB = PadlockButton(self)
        self._fieldOfViewLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._fieldOfViewLB)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 2, self._fieldOfViewLB)

        ## x flipped
        self._xFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._xFlippedQTWI.setText(0, "x flipped")
        self._xFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._xFlippedQTWI, 1, self._xFlippedCB)
        self._editableWidgets.append(self._xFlippedCB)
        self._xFlippedLB = PadlockButton(self)
        self._xFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xFlippedLB)
        self._tree.setItemWidget(self._xFlippedQTWI, 2, self._xFlippedLB)
        ## y flipped
        self._yFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._yFlippedQTWI.setText(0, "y flipped")
        self._yFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._yFlippedQTWI, 1, self._yFlippedCB)
        self._editableWidgets.append(self._yFlippedCB)
        self._yFlippedLB = PadlockButton(self)
        self._yFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._yFlippedLB)
        self._tree.setItemWidget(self._yFlippedQTWI, 2, self._yFlippedLB)
        # 2: sample
        self._sampleQTWI = qt.QTreeWidgetItem(self._tree)
        self._sampleQTWI.setText(0, "sample")
        ## x translation
        self._xTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._xTranslationQTWI.setText(0, "x translation")
        self._xTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._xTranslationQTWI, 1, self._xTranslationQLE)
        self._editableWidgets.append(self._xTranslationQLE)

        ## z translation
        self._zTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._zTranslationQTWI.setText(0, "z translation")
        self._zTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._zTranslationQTWI, 1, self._zTranslationQLE)
        self._editableWidgets.append(self._zTranslationQLE)

        # set up
        self._instrumentQTWI.setExpanded(True)
        self._sampleQTWI.setExpanded(True)
        self._beamQTWI.setExpanded(True)
        self._detectorQTWI.setExpanded(True)
        self.hideLockers(hide_lockers)

        # connect signal / slot
        self._energyEntry.editingFinished.connect(self._editingFinished)
        self._energyLockerLB.toggled.connect(self._editingFinished)
        self._xPixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._xPixelSizeLB.toggled.connect(self._editingFinished)
        self._yPixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._yPixelSizeLB.toggled.connect(self._editingFinished)
        self._distanceMetricEntry.editingFinished.connect(self._editingFinished)
        self._distanceLB.toggled.connect(self._editingFinished)
        self._fieldOfViewCB.currentIndexChanged.connect(self._editingFinished)
        self._fieldOfViewLB.toggled.connect(self._editingFinished)
        self._xFlippedCB.toggled.connect(self._editingFinished)
        self._xFlippedLB.toggled.connect(self._editingFinished)
        self._yFlippedCB.toggled.connect(self._editingFinished)
        self._yFlippedLB.toggled.connect(self._editingFinished)

    def update_tree(self) -> None:
        if self.getScan() is not None:
            self._updateInstrument()
            self._updateSample()
            self._tree.resizeColumnToContents(0)

    def _updateInstrument(self) -> None:
        scan = self.getScan()
        if scan is None:
            return
        else:
            for name, fct in {
                "energy": self._updateEnergy,
                "pixel size": self._updatePixelSize,
                "frame flips": self._updateFlipped,
                "field of view": self._updateFieldOfView,
                "sample-detector distance": self._updateDistance,
            }.items():
                try:
                    fct(scan=scan)
                except Exception as e:
                    _logger.error(f"Failed to update {name}. Error is {e}")

    def _updateSample(self) -> None:
        scan = self.getScan()
        if scan is None:
            return
        else:
            try:
                self._updateTranslations(scan=scan)
            except Exception as e:
                _logger.error(f"Fail to update translations. Error is {e}")

    def _updateTranslations(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)

        # note: for now and in order to allow edition we expect to have at most a unique value. Will fail for helicoidal
        def reduce(values):
            if values is None:
                return None
            values = numpy.array(values)
            values = numpy.unique(
                values[scan.image_key_control == ImageKey.PROJECTION.value]
            )
            if values.size == 1:
                return values[0]
            elif values.size == 0:
                return None
            else:
                return f"{values[0]} ... {values[-1]}"

        x_translation = reduce(scan.x_translation)
        z_translation = reduce(scan.z_translation)
        self._xTranslationQLE.setValue(x_translation)
        self._zTranslationQLE.setValue(z_translation)

    def _updateFieldOfView(self, scan: NXtomoScan) -> None:
        if not self._fieldOfViewLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            idx = self._fieldOfViewCB.findText(FOV.from_value(scan.field_of_view).value)
            if idx > 0:
                self._fieldOfViewCB.setCurrentIndex(idx)

    def _updateFlipped(self, scan: NXtomoScan) -> None:
        transformations = list(scan.get_detector_transformations(tuple()))
        transformation_matrix_det_space = build_matrix(transformations)
        if transformation_matrix_det_space is None or numpy.allclose(
            transformation_matrix_det_space, numpy.identity(3)
        ):
            flip_ud = False
            flip_lr = False
        elif numpy.array_equal(
            transformation_matrix_det_space, NXtomoEditorTask.Y_FLIP_MATRIX
        ):
            flip_ud = True
            flip_lr = False
        elif numpy.allclose(
            transformation_matrix_det_space, NXtomoEditorTask.Z_FLIP_MATRIX
        ):
            flip_ud = False
            flip_lr = True
        elif numpy.allclose(
            transformation_matrix_det_space, NXtomoEditorTask.Y_AND_Z_flip_MATRIX
        ):
            flip_ud = True
            flip_lr = True
        else:
            flip_ud = None
            flip_lr = None
            _logger.warning(
                "detector transformations provided not handled... For now only handle up-down flip as left-right flip"
            )
        if (not self._xFlippedLB.isLocked()) and flip_lr is not None:
            self._xFlippedCB.setChecked(flip_lr)
        if (not self._yFlippedLB.isLocked()) and flip_ud is not None:
            self._yFlippedCB.setChecked(flip_ud)

    def _updateDistance(self, scan: NXtomoScan) -> None:
        if not self._distanceLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._distanceMetricEntry.setValue(scan.distance)

    def _updateEnergy(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)
        if not self._energyLockerLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            energy = scan.energy
            self._energyEntry.setValue(energy)

    def _updatePixelSize(self, scan: NXtomoScan) -> None:
        assert isinstance(scan, NXtomoScan)
        if not self._xPixelSizeLB.isLocked():
            x_pixel_size = scan.x_pixel_size
            self._xPixelSizeMetricEntry.setValue(x_pixel_size)
        if not self._yPixelSizeLB.isLocked():
            y_pixel_size = scan.y_pixel_size
            self._yPixelSizeMetricEntry.setValue(y_pixel_size)

    def _editingFinished(self, *args, **kwargs):
        self.sigEditingFinished.emit()

    def hasLockField(self) -> bool:
        """return True if the widget has at least one lock field"""
        return True in [locker.isLocked() for locker in self._lockerPBs]

    def hideLockers(self, hide: bool) -> None:
        for locker in self._lockerPBs:
            locker.setVisible(not hide)

    def getEditableWidgets(self) -> tuple[qt.QWidget]:
        return tuple(self._editableWidgets)

    def setScan(self, scan: NXtomoScan | None) -> None:
        if scan is None:
            self._scan = scan
        elif not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"{scan} is expected to be an instance of {NXtomoScan}. Not {type(scan)}"
            )
        else:
            self._scan = weakref.ref(scan)
        self._scanInfoQLE.setScan(scan)
        # scan will only be read and not kept
        self.update_tree()

    def getScan(self) -> NXtomoScan | None:
        if self._scan is None or self._scan() is None:
            return None
        else:
            return self._scan()

    def getConfiguration(self) -> dict:
        """
        Return a dict with field full name as key
        and a tuple as value (field_value, is_locked)

        limitation: for now sample position are not handled because this is a 'corner case' for now
        """
        return {
            NXtomoEditorKeys.ENERGY: (
                self._energyEntry.getValue(),
                self._energyLockerLB.isLocked(),
            ),
            NXtomoEditorKeys.X_PIXEL_SIZE: (
                self._xPixelSizeMetricEntry.getValue(),
                self._xPixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.Y_PIXEL_SIZE: (
                self._yPixelSizeMetricEntry.getValue(),
                self._yPixelSizeLB.isLocked(),
            ),
            NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE: (
                self._distanceMetricEntry.getValue(),
                self._distanceLB.isLocked(),
            ),
            NXtomoEditorKeys.FIELD_OF_VIEW: (
                self._fieldOfViewCB.currentText(),
                self._fieldOfViewLB.isChecked(),
            ),
            NXtomoEditorKeys.X_FLIPPED: (
                self._xFlippedCB.isChecked(),
                self._xFlippedLB.isChecked(),
            ),
            NXtomoEditorKeys.Y_FLIPPED: (
                self._yFlippedCB.isChecked(),
                self._yFlippedLB.isChecked(),
            ),
            NXtomoEditorKeys.X_TRANSLATION: (self._xTranslationQLE.getValue(),),
            NXtomoEditorKeys.Z_TRANSLATION: (self._zTranslationQLE.getValue(),),
        }

    def setConfiguration(self, config: dict) -> None:
        energy = config.get("instrument.beam.energy", None)
        if energy is not None:
            energy, energy_locked = energy
            self._energyEntry.setValue(energy)
            self._energyLockerLB.setLock(energy_locked)

        x_pixel_size = config.get("instrument.detector.x_pixel_size", None)
        if x_pixel_size is not None:
            x_pixel_size, x_pixel_size_locked = x_pixel_size
            self._xPixelSizeMetricEntry.setValue(x_pixel_size)
            self._xPixelSizeLB.setLock(x_pixel_size_locked)

        y_pixel_size = config.get("instrument.detector.y_pixel_size", None)
        if y_pixel_size is not None:
            y_pixel_size, y_pixel_size_locked = y_pixel_size
            self._yPixelSizeMetricEntry.setValue(y_pixel_size)
            self._yPixelSizeLB.setLock(y_pixel_size_locked)

        detector_sample_distance = config.get("instrument.detector.distance", None)
        if detector_sample_distance is not None:
            detector_sample_distance, distance_locked = detector_sample_distance
            self._distanceMetricEntry.setValue(detector_sample_distance)
            self._distanceLB.setLock(x_pixel_size_locked)

        field_of_view = config.get("instrument.detector.field_of_view", None)
        if field_of_view is not None:
            field_of_view, field_of_view_locked = field_of_view
            self._fieldOfViewCB.setCurrentText(field_of_view)
            self._fieldOfViewLB.setLock(field_of_view_locked)

        x_flipped = config.get("instrument.detector.x_flipped", None)
        if x_flipped is not None:
            x_flipped, x_flipped_locked = x_flipped
            x_flipped = x_flipped in (True, "True", "true")
            self._xFlippedCB.setChecked(x_flipped)
            self._xFlippedLB.setLock(x_flipped_locked)

        y_flipped = config.get("instrument.detector.y_flipped", None)
        if y_flipped is not None:
            y_flipped, y_flipped_locked = y_flipped
            y_flipped = y_flipped in (True, "True", "true")
            self._yFlippedCB.setChecked(y_flipped)
            self._yFlippedLB.setLock(y_flipped_locked)

    def getConfigurationForTask(self) -> dict:
        """
        default configuration is stored as field: (field_value, filed_is_locked) when the task expects field_key: field_value
        Because we need to be able to reload settings of the LockButton.
        But the task doesn't care about it. She only want to know which field must be edited. So we need to filter the dict value.
        """
        return {key: value[0] for key, value in self.getConfiguration().items()}

    def clear(self) -> None:
        self._tree.clear()


class _TranslationMetricEntry(MetricEntry):
    """
    Widget to define a translation along one axis.

    The behavior is limited at the moment.
    * either the array contains a unique value on the array and a float is displayed
    * either the array contains several unique values and then '...' will be displayed. Users cannot provide an array in this case.
    """

    LOADED_ARRAY = "loaded array"

    class TranslationValidator(qt.QDoubleValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setNotation(qt.QDoubleValidator.ScientificNotation)

        def validate(self, a0: str, a1: int):
            if "..." in a0:
                return (qt.QDoubleValidator.Acceptable, a0, a1)
            else:
                return super().validate(a0, a1)

    def __init__(self, name, default_unit="m", parent=None):
        super().__init__(name, default_unit=default_unit, parent=parent)
        self._qlePixelSize.setValidator(self.TranslationValidator(self))

    def getValue(self) -> float:
        """

        :return: the value in meter
        """
        if "..." in self._qlePixelSize.text():
            # in this case this is the representation of an array, we don;t wan't to overwrite it
            return self.LOADED_ARRAY
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return float(self._qlePixelSize.text()) * self.getCurrentUnit()


class EnergyEntry(qt.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(MetricEntry.DoubleValidator())

    def setValue(self, a0):
        if a0 is None:
            a0 = "unknown"
        else:
            a0 = str(a0)
        super().setText(a0)

    def getValue(self) -> float | None:
        txt = self.text().replace(" ", "")
        if txt in ("unknown", ""):
            return None
        else:
            return float(txt)



# coding: utf-8
from __future__ import annotations


import logging
import os
import shutil
import tempfile
import time

import pytest
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core import settings
from tomwer.core.utils.lbsram import mock_low_memory, is_low_on_memory
from tomwer.core.process.control.scanvalidator import ScanValidatorP
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF
from tomwer.gui.utils.waiterthread import QWaiterThread
from tomwer.tests.utils import skip_gui_test

logging.disable(logging.INFO)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestScanValidator(TestCaseQt):
    """
    Simple test to make sure the timeout of data watcher is working properly
    """

    LOOP_MEM_RELEASER_DURATION = 0.2
    NB_SCANS = 2

    def tearDown(self):
        mock_low_memory(False)

    @classmethod
    def setUpClass(cls):
        settings.mock_lsbram(True)

        cls.scanValidator = ScanValidatorP(memoryReleaser=QWaiterThread(0.5))

        cls.scans = []

        for _ in range(cls.NB_SCANS):
            scanID = tempfile.mkdtemp()
            MockEDF.mockScan(scanID=scanID, nRadio=10, nRecons=2, nPagRecons=0, dim=10)
            cls.scans.append(scanID)

    @classmethod
    def tearDownClass(cls):
        settings.mock_lsbram(False)

        for f in cls.scans:
            if os.path.isdir(f):
                shutil.rmtree(f)

    @pytest.mark.skipif(
        (settings.isOnLbsram() and is_low_on_memory(settings.get_lbsram_path())),
        reason="Lbsram already overloaded",
    )
    def testMemoryReleaseLoop(self):
        """
        Make sure the internal loop of the scan validator is active if we are
        on lbsram.
        """

        def add_all_scans():
            for scan in self.scans:
                self.scanValidator.addScan(EDFTomoScan(scan))
            self.assertEqual(len(self.scanValidator._scans), self.NB_SCANS)

        def my_wait():
            for _ in range(3):
                while self.qapp.hasPendingEvents():
                    self.qapp.processEvents()
                time.sleep(self.LOOP_MEM_RELEASER_DURATION * 2)

        add_all_scans()
        mock_low_memory(True)
        my_wait()
        self.assertEqual(len(self.scanValidator._scans), 0)

        mock_low_memory(False)
        add_all_scans()
        my_wait()
        self.assertEqual(len(self.scanValidator._scans), self.NB_SCANS)
        self.scanValidator.clear()
        my_wait()
        self.assertEqual(len(self.scanValidator._scans), 0)



from __future__ import annotations
from silx.gui import qt

from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.gui.utils.gpu import SelectGPUGroupBox


class NabuPlatformSettings(qt.QWidget, _NabuStageConfigBase):
    """define settings to be used for the local reconstruction like gpu..."""

    sigConfigChanged = qt.Signal()

    DEFAULT_GPU_FRACTION = 90
    DEFAULT_CPU_FRACTION = 90

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        qt.QWidget.__init__(self, parent, stage=None)
        _NabuStageConfigBase.__init__(self, stage=None)

        self.setLayout(qt.QFormLayout())
        self._gpuSelWidget = SelectGPUGroupBox()
        self._gpuSelWidget.setChecked(False)
        # keep default behavior: don't set any default gpu
        self.layout().addRow(self._gpuSelWidget)

        # gpu percentage
        self._GPUFractionLabel = qt.QLabel("GPU percentage usage", self)
        self.registerWidget(self._GPUFractionLabel, "advanced")
        self._GPUFractionSB = qt.QSpinBox(self)
        self._GPUFractionSB.setSingleStep(10)
        self._GPUFractionSB.setSuffix("%")
        self._GPUFractionSB.setMinimum(1)
        self._GPUFractionSB.setMaximum(100)
        self._GPUFractionSB.setToolTip("Which fraction of GPU memory to use.")
        self.registerWidget(self._GPUFractionSB, "advanced")
        self.layout().addRow(self._GPUFractionLabel, self._GPUFractionSB)

        # cpu percentage
        self._CPUFractionLabel = qt.QLabel("CPU percentage usage", self)
        self.registerWidget(self._CPUFractionLabel, "advanced")
        self._CPUFractionSB = qt.QSpinBox(self)
        self._CPUFractionSB.setSingleStep(10)
        self._CPUFractionSB.setSuffix("%")
        self._CPUFractionSB.setMinimum(1)
        self._CPUFractionSB.setMaximum(100)
        self._CPUFractionSB.setToolTip("Which fraction of CPU memory to use.")
        self.registerWidget(self._CPUFractionSB, "advanced")
        self.layout().addRow(self._CPUFractionLabel, self._CPUFractionSB)

        # set up
        self._GPUFractionSB.setValue(NabuPlatformSettings.DEFAULT_GPU_FRACTION)
        self._CPUFractionSB.setValue(NabuPlatformSettings.DEFAULT_CPU_FRACTION)

        # connect signal / slot
        self._gpuSelWidget.sigGPUChanged.connect(self.sigConfigChanged)
        self._GPUFractionSB.valueChanged.connect(self._triggerSigConfChanged)
        self._CPUFractionSB.valueChanged.connect(self._triggerSigConfChanged)

    def getCPUFraction(self) -> float:
        return self._CPUFractionSB.value() / 100.0

    def setCPUFraction(self, value: float):
        return self._CPUFractionSB.setValue(int(value * 100.0))

    def getGPUFraction(self) -> float:
        return self._GPUFractionSB.value() / 100.0

    def setGPUFraction(self, value: float):
        return self._GPUFractionSB.setValue(int(value * 100.0))

    def _triggerSigConfChanged(self, *args, **kwargs) -> None:
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        config = {
            "gpu_mem_fraction": self.getGPUFraction(),
            "cpu_mem_fraction": self.getCPUFraction(),
        }
        selected_gpu_id = self._gpuSelWidget.getSelectedGPUId()
        if selected_gpu_id is not None:
            config["resources"] = {
                "gpu_id": selected_gpu_id,
            }
        return config

    def setConfiguration(self, config: dict) -> None:
        gpu_id = config.get("resources", {}).get("gpu_id", None)
        if gpu_id not in (None, ""):
            self._gpuSelWidget.setSelectedGPUId(value=gpu_id)
            self._gpuSelWidget.setChecked(True)
        if "gpu_mem_fraction" in config:
            self.setGPUFraction(config["gpu_mem_fraction"])
        if "cpu_mem_fraction" in config:
            self.setCPUFraction(config["cpu_mem_fraction"])



# from __future__ import annotations

# import logging

# from orangewidget import gui
# from orangewidget.settings import Setting
# from orangewidget.widget import Input, Output
# from orangecontrib.tomwer.orange.managedprocess import TomwerWithStackStack
# from silx.gui import qt
# from typing import Optional

# from tomwer.core.futureobject import FutureTomwerObject
# from tomwer.core.scan.scanbase import TomwerScanBase
# from tomwer.gui.reconstruction.nabu.helical import HelicalPrepareWeightsDouble
# from tomwer.core.process.reconstruction.nabu.helical import (
#     NabuHelicalPrepareWeightsDouble,
# )
# from tomwer.core.scan.nxtomoscan import NXtomoScan

# _logger = logging.getLogger(__name__)


# class NabuHelicalPrepareWeightsDoubleOW(
#     TomwerWithStackStack,
#     ewokstaskclass=NabuHelicalPrepareWeightsDouble,
# ):
#     """
#     widget used to call the `nabu-helical-prepare-weights-double` application on a dedicated thread. It define weights map and double flat field.

#     :param parent: the parent widget
#     """

#     # note of this widget should be the one registered on the documentation
#     name = "helical prerate weights double"
#     id = "orangecontrib.tomwer.widgets.reconstruction.NabuHelicalPrepareWeightsDoubleOW.NabuHelicalPrepareWeightsDoubleOW"
#     description = "compute map of weights requested for nabu helical reconstruction"
#     icon = "icons/nabu_prepare_weights_double.svg"
#     priority = 199
#     keywords = [
#         "tomography",
#         "nabu",
#         "reconstruction",
#         "nabu-helical",
#         "helical",
#         "weights",
#         "prepare",
#     ]

#     want_main_area = True
#     want_control_area = False
#     resizing_enabled = True

#     _ewoks_default_inputs = Setting(
#         {
#             "data": None,
#             "transition_width_vertical": 50,
#             "transition_width_horizontal": 50,
#             "processes_file": "",
#             "rotation_axis_position": 0,
#         }
#     )

#     _ewoks_inputs_to_hide_from_orange = (
#         "progress",
#         "processes_file",
#         "transition_width_vertical",
#         "transition_width_horizontal",
#         "rotation_axis_position",
#     )

#     sigScanReady = qt.Signal(TomwerScanBase)
#     "Signal emitted when a scan is ended"

#     TIMEOUT = 30

#     class Inputs:
#         data = Input(
#             name="data",
#             type=TomwerScanBase,
#             doc="one scan to be process",
#             default=True,
#             multiple=False,
#         )

#     class Outputs:
#         data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
#         future_tomo_obj = Output(
#             name="future_tomo_obj",
#             type=FutureTomwerObject,
#             doc="future object (process remotely)",
#         )

#     LOGGER = _logger

#     def __init__(self, parent=None):
#         super().__init__(parent)
#         self.__scan = None

#         # gui definition
#         _layout = gui.vBox(self.mainArea, self.name).layout()
#         self.widget = HelicalPrepareWeightsDouble(parent=self)
#         _layout.addWidget(self.widget)

#         ## connect signal / slot
#         self.widget.sigConfigChanged.connect(self._updateSettings)

#         if isinstance(self.task_output_changed_callbacks, set):
#             self.task_output_changed_callbacks.add(self._notify_state)
#         elif isinstance(self.task_output_changed_callbacks, list):
#             self.task_output_changed_callbacks.append(self._notify_state)
#         else:
#             raise NotImplementedError

#         ## handle settings
#         self._loadSettings()
#         self.task_executor_queue.sigComputationStarted.connect(self._newTaskStarted)

#     def _updateSettings(self):
#         config = self.widget.getConfiguration()
#         for key in ("transition_width", "processes_file"):
#             self._ewoks_default_inputs[key] = config[key]  # pylint: disable=E1137

#     @property
#     def request_input(self):
#         return self.__request_input

#     @request_input.setter
#     def request_input(self, request):
#         self.__request_input = request

#     def get_task_inputs(self):
#         assert self.__scan is not None
#         return {
#             "data": self.__scan,
#             "transition_width": self.widget.getConfiguration()["transition_width"],
#             "processes_file": self.widget.getConfiguration()["processes_file"],
#         }

#     def handleNewSignals(self) -> None:
#         """Invoked by the workflow signal propagation manager after all
#         signals handlers have been called.
#         """
#         # for now we want to avoid propagation any processing.
#         # task will be executed only when the user validates the dialog
#         data = super().get_task_inputs().get("data", None)
#         if data is not None:
#             if not isinstance(data, NXtomoScan):
#                 raise TypeError(
#                     f"data is expected to be an instance of NXtomoScan. {type(data)} are not handled"
#                 )
#             self.add(data.path)

#     def _loadSettings(self):
#         self.widget.setConfiguration(self._ewoks_default_inputs)

#     def _newTaskStarted(self):
#         try:
#             task_executor = self.sender()
#             scan = task_executor.current_task.inputs.data
#             self.notify_on_going(scan)
#         except Exception:
#             pass

# def _notify_state(self):
#     try:
#         task_executor = self.sender()
#         task_suceeded = task_executor.succeeded
#         scan = task_executor.current_task.outputs.data
#         if task_suceeded:
#             self.notify_succeed(scan=scan)
#         else:
#             self.notify_failed(scan=scan)
#     except Exception as e:
#         _logger.error(f"failed to handle task finished callback. Reason is {e}")

# @Inputs.data
# def process_data(self, scan: TomwerScanBase | None):
#     if scan is None:
#         return
#     else:
#         self.__scan = scan
#         self.notify_pending(scan=scan)
#         self.execute_ewoks_task()

#     def sizeHint(self) -> qt.QSize:
#         return qt.QSize(650, 60)



# coding: utf-8
from __future__ import annotations


import os
import weakref
from collections import OrderedDict

from silx.gui import qt

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from datetime import datetime


class ScanObservation(qt.QWidget):
    """
    Widget displayed the on-going observations (done by the data-listener)

    For each scan on-going we can display the following information:
    * time (when the scan has been discovered)
    * type: HDF5 or EDF
    * N projections: number of projections already acquired (a)
    * status of the scan (starting, on-going...)
    * acquisition: id of the scan (entry and data path in case of HDF5 scan)

    (a) the GUI was set up to display this information. Nevertheless this information is not provided by the rpc-call so this was never used.

    Context: when the 'data listener' gets triggered by bliss-tomo then we want to notify the user that a new scan is on-going.
    """

    HEADER = ("Time", "Type", "N Projections", "Status", "Acquisition")

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self._onGoingObservations = None
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(qt.QLabel(""))

        self.observationTable = qt.QTableView(parent=parent)
        self.observationTable.setSelectionBehavior(qt.QAbstractItemView.SelectRows)
        self.observationTable.setModel(
            _ObservedScanModel(parent=self.observationTable, header=self.HEADER)
        )
        self.observationTable.resizeColumnsToContents()
        self.observationTable.setSortingEnabled(True)
        self.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)

        self.layout().addWidget(self.observationTable)
        header = self.observationTable.horizontalHeader()
        header.setSectionResizeMode(0, qt.QHeaderView.Fixed)
        header.setSectionResizeMode(1, qt.QHeaderView.Fixed)
        header.setSectionResizeMode(2, qt.QHeaderView.Fixed)
        header.setSectionResizeMode(3, qt.QHeaderView.Fixed)
        header.setSectionResizeMode(4, qt.QHeaderView.Stretch)
        header.setStretchLastSection(True)

        self.observationTable.setColumnWidth(0, 70)
        self.observationTable.setColumnWidth(1, 40)
        self.observationTable.setColumnWidth(2, 90)
        self.observationTable.setColumnWidth(3, 70)

    @property
    def onGoingObservations(self):
        if self._onGoingObservations:
            return self._onGoingObservations()
        else:
            return None

    def setOnGoingObservations(self, onGoingObservations):
        """
        will update the table to display the observations contained in
        onGoingObservations

        :param onGoingObservations: the obsevations observed to display
        """
        if self.onGoingObservations:
            self.onGoingObservations.sigObsAdded.disconnect(self.addObservation)
            self.onGoingObservations.sigObsRemoved.disconnect(self.removeObservation)
            self.onGoingObservations.sigObsStatusReceived.disconnect(self.update)

        self._onGoingObservations = weakref.ref(onGoingObservations)
        self.onGoingObservations.sigObsAdded.connect(self.addObservation)
        self.onGoingObservations.sigObsRemoved.connect(self.removeObservation)
        self.onGoingObservations.sigObsStatusReceived.connect(self.update)

    def update(self, scan, status):
        """

        :param scan: the updated scan
        :param status: the status of the updated scan
        """
        self.observationTable.model().update(scan, status)

    def addObservation(self, scan):
        """

        :param scan: the scan observed
        """
        self.observationTable.model().add(scan, "starting")

    def removeObservation(self, scan):
        """

        :param scan: the scan removed
        """
        self.observationTable.model().remove(scan)

    def clear(self):
        self.observationTable.model().clear()


class _ObservedScanModel(qt.QAbstractTableModel):
    def __init__(self, parent, header, *args):
        qt.QAbstractTableModel.__init__(self, parent, *args)
        self.header = header
        self.observations = OrderedDict()
        self._time_stamps = OrderedDict()
        # note: the time stamp is the time of the discovery and not of the start scan (as we don't have this information)

    def add(self, scan, status):
        self.observations[scan] = status
        self._time_stamps[scan] = datetime.now()
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def remove(self, scan):
        self.observations.pop(scan, None)
        self._time_stamps.pop(scan, None)
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def update(self, scan, status):
        self.observations[scan] = status
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def clear(self):
        self.observations = OrderedDict()
        self._time_stamps = OrderedDict()
        if qt.qVersion() > "4.6":
            self.endResetModel()

    def rowCount(self, parent=None):
        return len(self.observations)

    def columnCount(self, parent=None):
        return len(self.header)

    def sort(self, col, order):
        self.layoutAboutToBeChanged.emit()
        if self.observations is None:
            return

        to_order = {}
        for observation in self.observations.keys():
            to_order[str(observation)] = observation

        ordering = sorted(list(to_order.keys()))
        if order == qt.Qt.DescendingOrder:
            ordering = reversed(ordering)
        _observations = OrderedDict()
        _time_stamps = OrderedDict()
        for str_key in ordering:
            key = to_order[str_key]
            _observations[key] = self.observations[key]
            _time_stamps[key] = self._time_stamps[key]

        self.observations = _observations
        self._time_stamps = _time_stamps
        self.layoutChanged.emit()

    def data(self, index, role):
        if index.isValid() is False:
            return None

        if role not in (qt.Qt.DisplayRole, qt.Qt.ToolTipRole):
            return None

        obs = list(self.observations.keys())[index.row()]
        time_stamp = self._time_stamps.get(obs, None)
        # time stamp
        if index.column() == 0:
            if time_stamp:
                return time_stamp.strftime("%H:%M:%S")
            else:
                return "Unknown timestamp"
        # acquisition type
        elif index.column() == 1:
            if isinstance(obs, TomwerScanBase):
                return obs.type
            elif isinstance(obs, BlissScan):
                return "hdf5"
            elif NXtomoScan.directory_contains_scan(directory=obs):
                return "hdf5"
            else:
                return "edf"
        # N projections
        elif index.column() == 2:
            if isinstance(obs, TomwerScanBase):
                return obs.tomo_n or 0
            elif isinstance(obs, BlissScan):
                return f"(at least) {obs.n_acquired or '?'} over {obs.tomo_n or '?'}"
            elif os.path.exists(obs) and os.path.isdir(obs):
                return str(len(os.listdir(obs)))
            else:
                return None
        elif index.column() == 3:
            return self.observations[obs]
        # observation id
        elif index.column() == 4:
            if role == qt.Qt.ToolTipRole:
                return obs
            elif isinstance(obs, (TomwerScanBase, BlissScan)):
                return str(obs)
            else:
                return os.path.basename(obs)
        else:
            return None

    def headerData(self, col, orientation, role):
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            return self.header[col]
        return None



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.dataportal.createscreenshots import CreateRawDataScreenshotsWidget


def test_CreateRawDataScreenshotsWidget(qtapp):  # noqa F811
    widget = CreateRawDataScreenshotsWidget()

    assert widget.getConfiguration() == {
        "raw_projections_required": True,
        "raw_projections_each": 90,
        "raw_darks_required": True,
        "raw_flats_required": True,
    }

    new_config = {
        "raw_projections_required": True,
        "raw_projections_each": 26,
        "raw_darks_required": False,
        "raw_flats_required": False,
    }
    widget.setConfiguration(new_config)

    assert widget.getConfiguration() == new_config



# coding: utf-8
from __future__ import annotations

import configparser
from collections import namedtuple

import numpy
import pytest
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from tomwer.core.process.reconstruction.nabu import settings as nabu_settings
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSlicesTask
from tomwer.core.process.reconstruction.normalization.normalization import (
    SinoNormalizationTask,
)
from tomwer.core.process.reconstruction.normalization.params import (
    SinoNormalizationParams,
)
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_SLICES,
)

try:
    import nabu
except ImportError:
    has_nabu = False
else:
    has_nabu = True
import os

_norm_setting = namedtuple("_norm_setting", ["method", "source", "extra_infos"])

_nabu_norm_field = namedtuple(
    "_nabu_norm_field",
    ["sino_normalization", "has_sino_normalization_file"],
)


normalization_config_test = (
    # no normalization
    (
        _norm_setting(method="none", source=None, extra_infos=None),
        _nabu_norm_field(
            sino_normalization="",
            has_sino_normalization_file=False,
        ),
    ),
    # chebyshev normalization
    (
        _norm_setting(method="chebyshev", source=None, extra_infos=None),
        _nabu_norm_field(
            sino_normalization="chebyshev",
            has_sino_normalization_file=False,
        ),
    ),
    # divide by a scalar
    (
        _norm_setting(method="division", source="scalar", extra_infos={"value": 12.0}),
        _nabu_norm_field(
            sino_normalization="division",
            has_sino_normalization_file=True,
        ),
    ),
    # substract a roi
    (
        _norm_setting(
            method="subtraction",
            source="manual ROI",
            extra_infos={
                "start_x": 0.0,
                "end_x": 1.0,
                "start_y": 0.0,
                "end_y": 1.0,
                "calc_fct": "mean",
                "calc_area": "volume",
                "calc_method": "scalar",
            },
        ),
        _nabu_norm_field(
            sino_normalization="subtraction",
            has_sino_normalization_file=True,
        ),
    ),
    # substract from a dataset
    (
        _norm_setting(
            method="subtraction",
            source="from dataset",
            extra_infos={
                "dataset_url": DataUrl(
                    file_path="random_dataset.hdf5",
                    data_path="data",
                    scheme="silx",
                ),
                "calc_fct": "median",
                "calc_area": "volume",
                "calc_method": "scalar",
            },
        ),
        _nabu_norm_field(
            sino_normalization="subtraction",
            has_sino_normalization_file=True,
        ),
    ),
)


@pytest.mark.skipif(
    (not has_nabu or nabu.version < "2022.2"),
    reason="nabu not available or the current version doesn't handle sino normlaization yet",
)
@pytest.mark.parametrize(
    "norm_setting, expected_nabu_conf", [item for item in normalization_config_test]
)
def test_normalization(norm_setting, expected_nabu_conf, tmp_path):
    """
    Insure normalization is correctly provided to nabu configuration file
    For this run a normalization process followed by a nabu process.
    """
    scan_dir = tmp_path / "scan"
    scan_dir.mkdir()

    nabu_cfg_folders = os.path.join(scan_dir, nabu_settings.NABU_CFG_FILE_FOLDER)
    os.makedirs(nabu_cfg_folders, exist_ok=True)

    random_dataset_file = os.path.join(nabu_cfg_folders, "random_dataset.hdf5")

    # create a random dataset if necessary
    with HDF5File(random_dataset_file, mode="w") as h5f:
        h5f["data"] = numpy.ones((20, 20))

    mock = MockNXtomo(
        scan_path=scan_dir,
        n_proj=10,
        n_ini_proj=10,
        scan_range=180,
        dim=20,
    )
    scan = mock.scan

    cfg_folder = os.path.join(
        str(scan_dir), PROCESS_FOLDER_RECONSTRUCTED_SLICES, "nabu_cfg_files"
    )
    cfg_file = os.path.join(cfg_folder, "entry_scan.cfg")
    assert not os.path.exists(cfg_file)
    norm_params = SinoNormalizationParams(
        method=norm_setting.method,
        source=norm_setting.source,
        extra_infos=norm_setting.extra_infos,
    )

    normalization = SinoNormalizationTask(
        inputs={
            "data": scan,
            "configuration": norm_params.to_dict(),
            "serialize_output_data": False,
        },
        varinfo=None,
    )
    normalization.run()

    # insure the method is style valid
    assert scan.intensity_normalization.method.value == norm_setting.method
    assert (
        "tomwer_processing_res_code" in scan.intensity_normalization.get_extra_infos()
    )

    process = NabuSlicesTask(
        inputs={
            "data": scan,
            "dry_run": True,
            "nabu_params": get_default_nabu_config(nabu_fullfield_default_config),
            "serialize_output_data": False,
        },
        varinfo=None,
    )
    process.run()
    assert os.path.exists(cfg_file)

    configuration = configparser.ConfigParser(allow_no_value=True)
    configuration.read(cfg_file)

    preproc_section = configuration["preproc"]
    sino_normalization = preproc_section.get("sino_normalization", "")
    sino_normalization_file = preproc_section.get("sino_normalization_file", "")
    assert sino_normalization == expected_nabu_conf.sino_normalization
    if expected_nabu_conf.has_sino_normalization_file:
        url = DataUrl(path=sino_normalization_file)
        assert url.is_valid()
    else:
        assert sino_normalization_file == ""
    assert (
        "tomwer_processing_res_code" in scan.intensity_normalization.get_extra_infos()
    )



from __future__ import annotations

from nabu.cuda.utils import collect_cuda_gpus
from silx.gui import qt


class SelectGPUGroupBox(qt.QGroupBox):
    """Interface to select a (local) GPU"""

    sigGPUChanged = qt.Signal()
    """Emit when the selected GPU changed"""

    def __init__(self, parent=None, title="select (local) GPU"):
        super().__init__(parent=parent, title=title)
        self.setCheckable(True)
        self.setLayout(qt.QFormLayout())
        self.__cuda_gpus = collect_cuda_gpus() or {}

        # warn user if no gpu found: reconstruction will work but will take way more time
        if len(self.__cuda_gpus) == 0:
            style = qt.QApplication.style()
            icon = style.standardIcon(qt.QStyle.SP_MessageBoxWarning)
            self.__iconLabel = qt.QLabel(parent=self)
            self.__iconLabel.setPixmap(icon.pixmap(80, 80))
            self.__noGPULabel = qt.QLabel("No cuda GPU detected")
            self.__noGPULabel.setToolTip(
                "No cuda GPU device detected. Reconstruction will be done on CPU which is much more slower"
            )
            self.layout().addRow(self.__iconLabel, self.__noGPULabel)

        self._gpusCB = qt.QComboBox(self)
        for gpu_id, gpu_desc in self.__cuda_gpus.items():
            name = gpu_desc.get("name", f"unknown gpu {gpu_id}")
            self._gpusCB.addItem(f"{name} (id-{gpu_id})", gpu_id)
        self.layout().addRow(self._gpusCB)

        # connect signal / slot
        self._gpusCB.currentIndexChanged.connect(self.__GPUChanged)

    @property
    def cuda_gpus(self) -> dict:
        return self.__cudaGPUs

    def getSelectedGPUId(self) -> int | None:
        if self.isChecked():
            selected_gpu = self._gpusCB.currentIndex()
            if selected_gpu == -1:
                return None
            else:
                return self._gpusCB.itemData(selected_gpu, qt.Qt.UserRole)
        else:
            return None

    def setSelectedGPUId(self, value: int):
        cb_index = self._gpusCB.findData(value, role=qt.Qt.UserRole)
        if cb_index >= 0:
            self._gpusCB.setCurrentIndex(cb_index)

    def __GPUChanged(self, *args, **kwargs):
        self.sigGPUChanged.emit()



"""widgets to perform a z-sttiching in pre processing (from projections) or post processing (from reconstructed volumes)"""

from __future__ import annotations


import logging

from nabu.stitching import config as _nabu_stitching_config
from nabu.stitching.config import StitchingType
from nabu.stitching.overlap import OverlapStitchingStrategy
from nabu.stitching.utils import ShiftAlgorithm as _NabuShiftAlgorithm
from silx.gui import qt

_logger = logging.getLogger(__name__)


class Axis_N_Params(qt.QGroupBox):
    """
    widget to edit the option for image registration along a specific axe
    """

    def __init__(self, title, parent=None) -> None:
        super().__init__(title, parent)
        self.setLayout(qt.QFormLayout())

        # img registration method
        self._imageRegMethodCB = qt.QComboBox(self)
        for method in _NabuShiftAlgorithm.values():
            self._imageRegMethodCB.addItem(method)
        self._imageRegMethodCB.setToolTip(
            "method to use in order to affine positions provided by the user"
        )
        self.layout().addRow("image registration method", self._imageRegMethodCB)

        # set up
        self._imageRegMethodCB.setCurrentText(_NabuShiftAlgorithm.NONE.value)

    def getCurrentMethod(self):
        return _NabuShiftAlgorithm.from_value(self._imageRegMethodCB.currentText())

    def setCurrentMethod(self, method):
        self._imageRegMethodCB.setCurrentText(
            _NabuShiftAlgorithm.from_value(method).value
        )

    def getOptsLine(self) -> str:
        current_method = self.getCurrentMethod()
        line_ = f"{_nabu_stitching_config.KEY_IMG_REG_METHOD}={current_method.value}"
        return line_

    def setOptsLine(self, opt_line: str) -> None:
        opt_line = opt_line.replace(",", ";")
        opt_line = opt_line.replace(" ", "")
        opts = {}
        for key_value in filter(None, opt_line.split(";")):
            key, value = key_value.split("=")
            opts[key] = value

        window_size = opts.get(_nabu_stitching_config.KEY_WINDOW_SIZE, None)
        if window_size is not None:
            try:
                self.setWindowSize(int(window_size))
            except Exception as e:
                _logger.error(e)

        img_reg_exp_method = opts.get(_nabu_stitching_config.KEY_IMG_REG_METHOD, None)
        if img_reg_exp_method is not None:
            try:
                self.setCurrentMethod(img_reg_exp_method)
            except Exception as e:
                _logger.error(e)


class AutoRefineWidget(qt.QWidget):
    """
    widget grouping information not specific to objects (output file, stitching strategy...)
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setLayout(qt.QVBoxLayout())

        self._globalOpts = qt.QWidget(self)
        self.layout().addWidget(self._globalOpts)
        self._globalOpts.setLayout(qt.QFormLayout())

        # stitching strategy
        self._stitchingStrategyCG = qt.QComboBox(parent=self)
        for strategy in OverlapStitchingStrategy.values():
            self._stitchingStrategyCG.addItem(strategy)
        self._globalOpts.layout().addRow(
            "stitching strategy", self._stitchingStrategyCG
        )
        # slice for cross correlation
        self._sliceForCorrelation = qt.QLineEdit("middle", self)
        self._globalOpts.layout().addRow(
            "slice for cross correlation", self._sliceForCorrelation
        )
        self._overwritePB = qt.QCheckBox("overwrite", self)
        self._globalOpts.layout().addRow(self._overwritePB)

        # pre processing options
        self._preProcGroup = qt.QGroupBox("pre-processing option", self)
        self.layout().addWidget(self._preProcGroup)
        self._preProcGroup.setLayout(qt.QFormLayout())
        # TODO: check if the widget with output .nx file exists somewhere
        self._outputFile = qt.QLineEdit("", self)
        self._preProcGroup.layout().addRow("output nexus file", self._outputFile)
        self._outputDataPath = qt.QLineEdit("entry0000", self)
        self._preProcGroup.layout().addRow("output data path", self._outputDataPath)

        # post processing options
        self._postProcGroup = qt.QGroupBox("post-processing option", self)
        self._postProcGroup.setLayout(qt.QFormLayout())
        self.layout().addWidget(self._preProcGroup)
        self._outputVolumeIdentifier = qt.QLineEdit("", self)
        self._postProcGroup.layout().addRow(
            "output identifier", self._outputVolumeIdentifier
        )
        self.layout().addWidget(self._postProcGroup)
        # spacer
        self._spacer = qt.QWidget(parent=self)
        self._spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacer)

        # set up
        self._stitchingStrategyCG.setCurrentText(
            OverlapStitchingStrategy.COSINUS_WEIGHTS.value
        )

    def getConfiguration(self) -> dict:
        raise NotImplementedError

    def setConfiguration(self, config: dict):
        raise NotImplementedError

    def updateStitchingType(self, mode: StitchingType):
        mode = StitchingType.from_value(mode)
        self._preProcGroup.setVisible(mode is StitchingType.Z_PREPROC)
        self._postProcGroup.setVisible(mode is StitchingType.Z_POSTPROC)

    def getSliceForPreview(self) -> int | str:
        slice = self._sliceForPreview.text()
        try:
            return int(slice)
        except Exception:
            return slice


class ManualRefineWidget(qt.QWidget):
    pass


class _RefineEstimationTabWidget(qt.QTabWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._autoRefine = AutoRefineWidget(parent=self)
        self.addTab(self._autoRefine, "automatique")
        self._manualRefine = ManualRefineWidget(parent=self)
        self.addTab(self._manualRefine, "manual")


class _SliceGetter(qt.QLineEdit):
    class Validator(qt.QIntValidator):
        def validate(self, a0: str, a1: int):
            if a0 in ("first", "middle", "last"):
                return (qt.QValidator.Acceptable, a0, a1)
            elif a0 in (
                "f",
                "fi",
                "fir",
                "firs",
                "m",
                "mi",
                "mid",
                "midd",
                "middl",
                "l",
                "la",
                "las",
            ):
                # dummy hack to avoid letting users write non-sense.
                return (qt.QValidator.Intermediate, a0, a1)
            else:
                return super().validate(a0, a1)

    def __init__(self, contents, parent=None):
        super().__init__(contents, parent)
        self.setValidator(self.Validator(parent=self))


class RefineEstimationCtrlWidget(qt.QWidget):
    sigSlicePreviewChanged = qt.Signal(str)
    """emit when the slice preview has changed"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        # main widget
        self.setLayout(qt.QGridLayout())
        self._widget = _RefineEstimationTabWidget(parent=self)
        self.layout().addWidget(self._widget, 0, 0, 3, 3)
        # slice to be used for preview
        self._previewedSliceQLE = _SliceGetter("middle", parent=self)
        self.layout().addWidget(self._previewedSliceQLE, 4, 1, 2, 1)
        # auto update preview
        self._updateAutoPreviewCB = qt.QCheckBox("auto update preview", self)
        self.layout().addWidget(self._updateAutoPreviewCB, 5, 1, 1, 2)

    def isPreviewAuto(self):
        return self._updateAutoPreviewCB.isChecked()

    def setPreviewAuto(self, auto: bool):
        self._updateAutoPreviewCB.setChecked(auto)

    def getSliceForPreview(self) -> str:
        return self._previewedSliceQLE.text()

    def setSliceForPreview(self, slice: str):
        self._previewedSliceQLE.setText(str(slice))



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile
import unittest

import h5py

from tomwer.core.process.control.datalistener.rpcserver import _BaseDataListenerThread


class JSONRPCClient:
    """Simulate call from bliss"""

    def scan_started(self, scan_number):
        return {
            "method": "scan_started",
            "params": [scan_number],
            "jsonrpc": "2.0",
            "id": 0,
        }

    def scan_ended(self, scan_number):
        return {
            "method": "scan_ended",
            "params": [scan_number],
            "jsonrpc": "2.0",
            "id": 1,
        }

    def sequence_started(
        self, saving_file, scan_title, sequence_scan_number, proposal_file, sample_file
    ):
        params = [
            saving_file,
            scan_title,
            sequence_scan_number,
            proposal_file,
            sample_file,
        ]
        return {
            "method": "sequence_started",
            "params": params,
            "jsonrpc": "2.0",
            "id": 2,
        }

    def sequence_ended(self, saving_file, sequence_scan_number, success):
        return {
            "method": "sequence_ended",
            "params": [saving_file, sequence_scan_number, success],
            "jsonrpc": "2.0",
            "id": 3,
        }


class TestJsonRPCServer(unittest.TestCase):
    """test the json rpc server"""

    def setUp(self):
        self.input_dir = tempfile.mkdtemp()
        sample_dir = os.path.join(self.input_dir, "sample")
        os.makedirs(sample_dir)
        self._proposal_file = os.path.join(sample_dir, "ihpayno_sample.h5")
        with h5py.File(self._proposal_file, mode="w") as h5f:
            h5f["test"] = "toto"

        os.makedirs(os.path.join(sample_dir, "sample_29042021"))
        self._sample_file = os.path.join(
            sample_dir, "sample_29042021", "sample_29042021.h5"
        )
        self._sample_file_entry = "1.1"

        with h5py.File(self._sample_file, mode="w") as h5f:
            h5f[self._sample_file_entry] = "tata"

        self.data_listener = _BaseDataListenerThread(host="localhost", port=4000)

    def tearDown(self):
        shutil.rmtree(self.input_dir)
        self.data_listener.stop()

    def testStartStop(self):
        self.data_listener.start()
        self.data_listener.stop()



# coding: utf-8
from __future__ import annotations

import logging

from ewokscore.task import Task as EwoksTask

from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.process.task import BaseProcessInfo
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.signal import Signal
from tomwer.core.utils import logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan

logger = logging.getLogger(__name__)


class _ScanValidatorPlaceHolder(EwoksTask, input_names=["data"], output_names=["data"]):
    def run(self):
        self.outputs.data = data_identifier_to_scan(self.inputs.data)


class ScanValidator(BaseProcessInfo):
    """
    Simple workflow locker until the use validate scans

    :param memReleaserWaitLoop: the time to wait in second between two
                                    memory overload if we are in lbsram.
    """

    def __init__(self, memoryReleaser):
        super().__init__()
        self._manualValidation = True
        self._hasToLimitScanBlock = settings.isOnLbsram()
        self._memoryReleaser = memoryReleaser
        self._scans = {}
        """associate scanId (keys) to :class:`.TomoBase` object (value)"""

        # add callback _loopMemoryReleaserto free memory if necessary on lbsram
        if self._hasToLimitScanBlock:
            self._memoryReleaser.finished.connect(self._loopMemoryReleaser)
            self._memoryReleaser.start()

    def __del__(self):
        self._clearMemoryReleaser()

    def _clearMemoryReleaser(self):
        if self._memoryReleaser is not None:
            self._memoryReleaser.should_be_stopped = True
            self._memoryReleaser.wait(4000)
            self._memoryReleaser = None

    def addScan(self, scan):
        """
        Return the index on the current orderred dict

        :param scan:
        :return:
        """
        _ftserie = scan
        if type(scan) is str:
            _ftserie = ScanFactory.create_scan_object(_ftserie)
        logger.info(f"Scan {_ftserie} has been added by the Scan validator")

        self._scans[str(_ftserie)] = _ftserie

        index = len(self._scans) - 1

        self._freeStackIfNeeded()
        return index

    def _freeStackIfNeeded(self):
        # if we are low in memory in lbsram: we will automatically validate the current scan
        isLowMemoryLbs = is_low_on_memory(settings.get_lbsram_path()) is True
        if not self.isValidationManual():
            self._validateStack()
        elif isLowMemoryLbs:
            mess = "low memory, free ScanValidator stack "
            logger.processSkipped(mess)
            self._validateStack(filter_="lbsram")

    def _loopMemoryReleaser(self):
        """
        simple loop using the _memoryReleaser and calling the
        _freeStackIfNeeded function
        """
        self._freeStackIfNeeded()
        if self._memoryReleaser and not hasattr(
            self._memoryReleaser, "should_be_stopped"
        ):
            self._memoryReleaser.start()

    def _validateStack(self, filter_=None):
        """
        :param filter: can be None or 'lbsram' if lbsram will only free scan
                       located on lbsram.
        Validate all the scans in the stack.
        """
        for scanID in list(self._scans.keys()):
            scan = self._scans[scanID]
            if filter_ is None or settings.isOnLbsram(scan):
                self._validated(scan)

    def _validateScan(self, scan):
        """This will validate the ftserie currently displayed

        :warning: this will cancel the currently displayed reconstruction.
            But if we are validating a stack of ftserie make sure this is the
            correct one you want to validate.
            Execution order in this case is not insured.
        """
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            self._validated(scan)

    def _cancelScan(self, scan):
        """This will cancel the ftserie currently displayed

        :warning: this will cancel the currently displayed reconstruction.
            But if we are validating a stack of ftserie make sure this is the
            correct one you want to validate.
            Execution order in this case is not insured.
        """
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            self._canceled(scan)

    def _redoAcquisitionScan(self, scan):
        """This will emit a signal to request am acquisition for the current
        ftSerieReconstruction

        :warning: this will cancel the currently displayed reconstruction.
            But if we are validating a stack of ftserie make sure this is the
            correct one you want to validate.
            Execution order in this case is not insured.
        """
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            self._redoacquisition(scan)

    # ------ callbacks -------
    def _validated(self, scan):
        """Callback when the validate button is pushed"""
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            info = f"{str(scan)} has been validated"
            logger.processEnded(
                info,
                extra={
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: str(scan),
                },
            )
            self._sendScanReady(scan)
            # in some case the scan can not be in the _scans
            # (if the signal) come from an other window that 'scan to treat'
            if str(scan) in self._scans:
                del self._scans[str(scan)]

    def _canceled(self, scan):
        """Callback when the cancel button is pushed"""
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            info = "%s has been canceled" % str(scan)
            logger.processEnded(
                info,
                extra={
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: str(scan),
                },
            )
            self._sendScanCanceledAt(scan)
            # in some case the scan can not be in the _scans
            # (if the signal) come from an other window that 'scan to treat'
            if str(scan) in self._scans:
                del self._scans[str(scan)]
            self.clear()

    def _redoacquisition(self, ftserie):
        """Callback when the redo acquisition button is pushed"""
        raise NotImplementedError("_redoacquisition not implemented yet")

    def _changeReconsParam(self, ftserie):
        """Callback when the change reconstruction button is pushed"""
        if ftserie is None:
            return

        _ftserie = ftserie
        if type(ftserie) is str:
            _ftserie = ScanFactory.create_scan_object(_ftserie)

        if _ftserie.path in self._scans:
            del self._scans[str(_ftserie)]
        self._sendUpdateReconsParam(_TomwerBaseDock(tomo_instance=_ftserie))

    def setManualValidation(self, b):
        """if the validation mode is setted to manual then we will wait for
        user approval before validating. Otherwise each previous and next scan
        will be validated

        :paramean b: False if we want an automatic validation
        """
        self._manualValidation = b
        if not self.isValidationManual():
            self._validateStack()

    def isValidationManual(self):
        """

        :return: True if the validation is waiting for user interaction,
                 otherwise False
        """
        return self._manualValidation

    def _sendScanReady(self, scan):
        raise RuntimeError("ScanValidator is a pure virtual class.")

    def _sendScanCanceledAt(self, scan):
        raise RuntimeError("ScanValidator is a pure virtual class.")

    def _sendUpdateReconsParam(self, scan):
        raise RuntimeError("ScanValidator is a pure virtual class.")

    def clear(self):
        scans = list(self._scans.values())
        # need to be copy because cancel will update self._scans
        for scan in scans:
            self._cancelScan(scan=scan)


class ScanValidatorP(ScanValidator):
    """
    For now to avoid multiple inheritance from QObject with the process widgets
    we have to define two classes. One only for the QObject inheritance.

    :param memReleaserWaitLoop: the time to wait in second between two
                                    memory overload if we are in lbsram.
    """

    scanReady = Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""
    scanCanceledAt = Signal(str)
    """Signal emitted when a scan has been canceled"""
    updateReconsParam = Signal(TomwerScanBase)
    """Signal emitted when a scan need to be reconstructed back with new
    parameters"""

    def __init__(self, memoryReleaser=None):
        ScanValidator.__init__(self, memoryReleaser)

    def _sendScanReady(self, scan):
        self.scanReady.emit(scan)

    def _sendScanCanceledAt(self, scan):
        self.scanCanceledAt.emit(scan)

    def _sendUpdateReconsParam(self, scan):
        self.updateReconsParam.emit(scan)



# coding: utf-8
from __future__ import annotations


import logging

from silx.gui import qt

from tomwer.core.process.reconstruction.darkref import params as dkrf

logger = logging.getLogger(__name__)


class QDKRFRP(dkrf.DKRFRP, qt.QObject):
    sigChanged = qt.Signal()
    """Signal emitted when at least one element of the dictionary change"""

    def __init__(self):
        qt.QObject.__init__(self)
        dkrf.DKRFRP.__init__(self)

    def changed(self):
        self.sigChanged.emit()



# coding: utf-8
from __future__ import annotations

import gc
import os
import shutil
import tempfile
import time

from orangecanvas.scheme.readwrite import literal_dumps
from processview.core.manager.manager import DatasetState, ProcessManager
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomoscan.esrf.scan.mock import MockEDF

from orangecontrib.tomwer.widgets.control.EDF2NXTomomillOW import EDF2NXOW
from orangecontrib.tomwer.widgets.control.NXTomomillOW import NXTomomillOW
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockBlissAcquisition


class TestNXTomomillOw(TestCaseQt):
    """
    Test the NXTomomillOW widget
    """

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        mock = MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_scan_per_sequence=3,
            n_darks=2,
            n_flats=2,
            output_dir=self.tempdir,
        )
        self.scan = BlissScan(
            master_file=mock.samples[0].sample_file,
            proposal_file=mock.proposal_file,
            entry="1.1",
        )
        self.widget = NXTomomillOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.tempdir)
        gc.collect()

    def test(self):
        self.widget.add(self.scan.master_file)
        self.widget._sendAll()

    def test_literal_dumps(self):
        self.widget._saveNXTomoCfgFile(cfg_file="")
        literal_dumps(self.widget._ewoks_default_inputs)


class TestEDF2NXOW(TestCaseQt):
    """
    Test the EDF2NXOW widget
    """

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        self._edf_folder_1 = os.path.join(self.tempdir, "edf_scan_1")
        self._edf_folder_2 = os.path.join(self.tempdir, "edf_scan_2_")
        MockEDF(
            scan_path=self._edf_folder_1,
            n_radio=10,
            n_ini_radio=10,
            n_extra_radio=0,
            dim=128,
            dark_n=1,
            flat_n=1,
        )
        MockEDF(
            scan_path=self._edf_folder_2,
            n_radio=10,
            n_ini_radio=10,
            n_extra_radio=0,
            dim=128,
            dark_n=1,
            flat_n=1,
        )
        self.widget = EDF2NXOW()
        self._process_manager = ProcessManager()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.tempdir)
        gc.collect()

    def test(self):
        self.widget.add(self._edf_folder_1)
        self.widget.add(self._edf_folder_2)
        self.widget._sendAll()
        self.qapp.processEvents()
        timeout = 10
        processing_queue = self.widget.task_executor_queue
        while not processing_queue.is_available and not processing_queue.empty():
            self.qapp.processEvents()
            time.sleep(0.1)
            timeout -= 0.1
            if timeout <= 0:
                raise TimeoutError
        # need to call one more time the processEvents tpo call callbacks and status notification
        for i in range(3):
            self.qapp.processEvents()
            time.sleep(0.1)

        # check the two output files exists and status of the processing is done...
        for folder in (self._edf_folder_1, self._edf_folder_2):
            scan = NXtomoScan(scan=folder + ".nx", entry="entry")
            self.assertTrue(os.path.exists(scan.master_file))

            self.assertEqual(
                self._process_manager.get_dataset_state(
                    dataset_id=scan.get_identifier(),
                    process=self.widget,
                ),
                DatasetState.SUCCEED,
            )

    def test_literal_dumps(self):
        self.widget._saveNXTomoCfgFile(cfg_file="")
        literal_dumps(self.widget._ewoks_default_inputs)



# coding: utf-8
from __future__ import annotations

import copy
import functools
import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from processview.core import helpers as pv_helpers
from processview.core.manager import DatasetState
from processview.gui.processmanager import ProcessManager

from silx.gui import qt

import tomwer.core.process.reconstruction.axis
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.process.reconstruction.axis import AxisTask
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.gui.reconstruction.axis import AxisMainWindow
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.stacks.reconstruction.axis import AxisProcessStack

from ..utils import WidgetLongProcessing

logger = logging.getLogger(__name__)


class AxisOW(SuperviseOW, WidgetLongProcessing):
    """
    Widget used to defined the center of rotation axis to be used for a
    reconstruction.

    :param _connect_handler: True if we want to store the modifications
                      on the setting. Need for unit test since
                      keep alive qt widgets.
    """

    name = "center of rotation finder"
    id = "orange.widgets.tomwer.axis"
    description = "use to compute the center of rotation"
    icon = "icons/axis.png"
    priority = 14
    keywords = [
        "tomography",
        "axis",
        "tomwer",
        "reconstruction",
        "rotation",
        "position",
        "center of position",
    ]

    ewokstaskclass = tomwer.core.process.reconstruction.axis.AxisTask

    want_main_area = True
    resizing_enabled = True

    settingsHandler = CallbackSettingsHandler()

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    _ewoks_default_inputs = Setting({"data": None, "axis_params": None, "gui": None})

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            multiple=True,
        )
        data_recompute_axis = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="recompute delta / beta",
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None, axis_params: QAxisRP | None = None):
        """

        :param parent: QWidget parent or None
        :param _connect_handler: used for CI, because if connected fails CI
        :param axis_params: reconstruction parameters
        """
        if axis_params is not None:
            if not isinstance(axis_params, QAxisRP):
                raise TypeError(
                    f"axis_params should be an instance of QAxisRP. Not {type(axis_params)}"
                )
        self._axis_params = axis_params or QAxisRP()

        # handle settings
        #  axis params settings
        axis_params_settings = self._ewoks_default_inputs.get("axis_params", None)
        if axis_params_settings not in (None, dict()):
            try:
                self._axis_params.load_from_dict(axis_params_settings)
            except Exception as e:
                logger.error(f"fail to load reconstruction settings: {e}")

        #  gui settings
        gui_settings = self._ewoks_default_inputs.get("gui", {})
        if gui_settings is None:
            gui_settings = {}

        original_mode = self._axis_params.mode
        if original_mode is AxisMode.manual:
            original_cor = self._axis_params.relative_cor_value
        else:
            original_cor = None

        self.__lastAxisProcessParamsCache = None
        # used to memorize the last (reconstruction parameters, scan)
        self.__scan = None
        self.__skip_exec = False
        self._n_skip = 0
        self._patches = []
        """patches for processing"""

        WidgetLongProcessing.__init__(self)
        SuperviseOW.__init__(self, parent)
        self._processingStack = AxisProcessStack(
            axis_params=self._axis_params, process_id=self.process_id
        )

        self._widget = AxisMainWindow(parent=self, axis_params=self._axis_params)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._layout.addWidget(self._widget)

        # load settings
        try:
            if "mode_is_lock" in gui_settings:
                mode_lock = gui_settings["mode_is_lock"]
                # if the mode is manual or read ignore lock
                if not (
                    mode_lock is True
                    and self.getMode() in (AxisMode.manual, AxisMode.read)
                ):
                    self._setModeLockFrmSettings(mode_lock)

            if "value_is_lock" in gui_settings:
                if gui_settings["value_is_lock"] is True:
                    self._setValueLockFrmSettings(bool(gui_settings["value_is_lock"]))

            auto_update_estimated_cor = gui_settings.get(
                "auto_update_estimated_cor", True
            )
            self._widget.setAutoUpdateEstimatedCor(auto_update_estimated_cor)
            self._widget.setYAxisInverted(gui_settings.get("y_axis_inverted", False))

        except Exception as e:
            logger.warning(f"Fail to load settings. Error is {str(e)}")

        # expose API
        self._applyBut = self._widget._controlWidget._applyBut
        # connect Signal / Slot
        self._widget.sigComputationRequested.connect(self.__compute)
        self._widget.sigApply.connect(self.__validate)
        self._widget.sigAxisEditionLocked.connect(self.__lockReconsParams)
        self._processingStack.sigComputationStarted.connect(self._processingStart)
        self._processingStack.sigComputationEnded.connect(self._scanProcessed)

        self._axis_params.sigChanged.connect(self._updateSettingsVals)
        self._widget._axisWidget._settingsWidget._mainWidget._calculationWidget.sigUpdateXRotAxisPixelPosOnNewScan.connect(
            self._updateSettingsVals
        )
        self._widget.sigAxisEditionLocked.connect(self._updateSettingsVals)
        self._widget.sigModeChanged.connect(self._updateSettingsVals)
        self._widget.sigLockModeChanged.connect(self._updateSettingsVals)

        self._widget._axisWidget._settingsWidget._mainWidget._calculationWidget._estimatedCorWidget.sigValueChanged.connect(
            self._updateSettingsVals
        )

        # handle special case of the manual mode
        # force axis to manual because not handled by the widget setup directly
        if original_mode in (
            AxisMode.manual,
            AxisMode.read,
        ):
            self._axis_params.mode = original_mode
            if original_cor is not None:
                self._axis_params.set_relative_value(original_cor)

            self.setMode(original_mode)
            # force gui update
            self._widget.sigModeChanged.emit("manual")
            # hard fix, but this widget has to be redone and rework anyway
            self._widget._axisWidget._settingsWidget._mainWidget._calculationWidget._modeChanged()
            if original_cor is not None:
                self._widget._controlWidget._positionInfo.setPosition(
                    relative_cor=original_cor, abs_cor=None
                )

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def _processingStart(self, *args, **kwargs):
        WidgetLongProcessing._startProcessing(self)

    def _scanProcessed(self, scan, future_tomo_obj):
        assert isinstance(scan, TomwerScanBase)
        WidgetLongProcessing._endProcessing(self, scan)
        if self.isValueLock() or self.isModeLocked():
            self.__scan = scan
            self.__validate()
        else:
            pm = ProcessManager()
            # handle previous scan if not validated yey: skip it
            if (
                self.__scan is not None
                and pm.get_dataset_state(dataset=self.__scan, process=self)
                != DatasetState.SUCCEED
            ):
                pm.notify_dataset_state(
                    dataset=self.__scan,
                    process=self,
                    state=DatasetState.SKIPPED,
                    details="Scan was waiting for validation. Has been replaced by another scan. No scan stack on Axis process",
                )

            # handle the new scan
            # retrieve details to keep them in `memory`
            self.__scan = scan
            details = pm.get_dataset_details(dataset=self.__scan, process=self)
            details = "Wait for user validation. " + details
            pm.notify_dataset_state(
                dataset=self.__scan,
                process=self,
                state=DatasetState.WAIT_USER_VALIDATION,
                details=details,
            )
            self.activateWindow()
            self.raise_()
            self.show()

    def __compute(self):
        if self.__scan:
            dict_for_cache = self._axis_params.to_dict().copy()
            dict_for_cache.pop("POSITION_VALUE")
            params_cache = (
                dict_for_cache,
                str(self.__scan),
            )
            # check mode is not locked or manual and value is still here
            if (
                params_cache == self.__lastAxisProcessParamsCache
            ) and self._processingStack.is_computing():
                logger.error(
                    "Parameters and scan are the same as last request. Please wait until the processing is done."
                )
            else:
                self.__lastAxisProcessParamsCache = params_cache
                callback = functools.partial(self._updatePosition, self.__scan)
                self._processingStack.add(
                    data=self.__scan,
                    configuration=self._axis_params.to_dict(),
                    callback=callback,
                )

    def __validate(self):
        """Validate the current scan and move the scan to the next process.
        The Axis will process the next scan in the stack.
        """
        if self.__scan:
            if self.getMode() is AxisMode.manual:
                # when value is set from manual mode we need to retrieve it from the GUI `_axis_param`. Else it has already been set
                # to the scan `_axis_params`
                if (
                    self.__scan._axis_params.frame_width is None
                    and self.__scan.dim_1 is not None
                ):
                    self.__scan._axis_params.frame_width = self.__scan.dim_1

                self.__scan._axis_params.set_relative_value(
                    self._axis_params.relative_cor_value
                )
                _cor_value = self.__scan.axis_params.relative_cor_value
                pv_helpers.notify_succeed(
                    process=self,
                    dataset=self.__scan,
                    details=f"axis calculation defined for {self.__scan.path}: {_cor_value} (using manual)",
                )
            # validate the center of rotation
            pm = ProcessManager()
            # retrieve details to keep them in memory
            details = pm.get_dataset_details(dataset=self.__scan, process=self)
            state = pm.get_dataset_state(dataset=self.__scan, process=self)
            if state in (DatasetState.ON_GOING, DatasetState.WAIT_USER_VALIDATION):
                # update the state to SUCCEED
                ProcessManager().notify_dataset_state(
                    dataset=self.__scan,
                    process=self,
                    state=DatasetState.SUCCEED,
                    details=details,
                )
            self.accept()
            self.scan_ready(scan=self.__scan)
        self.hide()

    def __lockReconsParams(self, lock):
        self.lock_position_value(lock)

    def scan_ready(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)

    def _informNoProjFound(self, scan):
        msg = qt.QMessageBox(self)
        msg.setIcon(qt.QMessageBox.Warning)
        text = (
            "Unable to find url to compute the axis of `%s`" % scan.path
            or "no path given"
        )
        text += ", please select them from the `axis input` tab"
        msg.setText(text)
        msg.exec_()

    def _updateSettingsVals(self):
        # remove rp setting to the advantage of 'static_input'
        self._ewoks_default_inputs = {
            "data": None,
            "axis_params": self._axis_params.to_dict(),
            "gui": {
                "mode_is_lock": self.isModeLocked(),
                "value_is_lock": self.isValueLock(),
                "auto_update_estimated_cor": self._widget.getAutoUpdateEstimatedCor(),
                "y_axis_inverted": self._widget.isYAxisInverted(),
            },
        }

    def _skip_exec(self, b):
        """util function used for unit test. If activate, skip the call to
        self.exec() in process"""
        self.__skip_exec = b

    @property
    def recons_params(self):
        return self._axis_params

    def _lock_axis_controls(self, lock):
        """

        :param lock: lock the axis controls to avoid modification of the
                          requested options, method... of the axis calculation
                          when this value is under calculation.
        """
        self._widget.setLocked(lock)

    def isValueLock(self):
        """
        Check if the cor value has been lock. If so we simply copy the cor
        value and move to the next scan
        """
        return self._widget.isValueLock()

    def isModeLocked(self):
        """
        Check if the mode has been lock or not. If lock then call the
        algorithm and does not wait for any user feedback
        """
        return self._widget.isModeLock()

    @Inputs.data
    def new_data_in(self, scan, *args, **kwargs):
        if scan is None:
            return
        scan_ = copy.copy(scan)
        if not (
            settings.isOnLbsram(scan) and is_low_on_memory(settings.get_lbsram_path())
        ):
            set_position = False  # avoid shift reset
            self._widget.setScan(scan=scan_, set_position=set_position)
        elif scan_.axis_params is None:
            scan_.axis_params = QAxisRP()
        self.process(scan=scan_)

    def process(self, scan):
        if scan is None:
            return
        self.__scan = scan
        self._axis_params.frame_width = scan.dim_1
        if (
            settings.isOnLbsram(scan)
            and is_low_on_memory(settings.get_lbsram_path()) is True
        ):
            self._updatePosition(scan=scan)
            self.scan_ready(scan=scan)
        elif self.__skip_exec:
            self._n_skip += 1
            if self.isValueLock():
                scan._axis_params.set_relative_value(
                    self._axis_params.relative_cor_value
                )
                cor = scan._axis_params.relative_cor_value
                if isinstance(scan, NXtomoScan):
                    entry = scan.entry
                else:
                    entry = "entry"
                with scan.acquire_process_file_lock():
                    ap = AxisTask(
                        process_id=self.process_id,
                        inputs={
                            "data": None,
                        },
                    )
                    try:
                        AxisTask._register_process(
                            process_file=scan.process_file,
                            entry=entry,
                            process=ap,
                            results={
                                "center_of_rotation": cor if cor is not None else "-"
                            },
                            configuration=self._axis_params.to_dict(),
                            process_index=scan.pop_process_index(),
                            overwrite=True,
                        )
                    except Exception as e:
                        logger.warning(f"Fail to register Axis process. Error is {e}")
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=ap,
                        state=DatasetState.SUCCEED,
                    )
            else:
                processing_class = AxisTask(
                    inputs={
                        "axis_params": self._axis_params,
                        "data": self.__scan,
                        "serialize_output_data": False,
                    }
                )
                processing_class.run()
            self._updatePosition(scan=self.__scan)
            if self.isModeLocked() or self.isValueLock():
                self.scan_ready(scan=scan)

        elif self.isValueLock():
            entry = scan.entry if isinstance(scan, NXtomoScan) else "entry"
            cor = self._axis_params.relative_cor_value
            scan.axis_params.set_relative_value(cor)
            scan._axis_params.mode = "manual"
            with scan.acquire_process_file_lock():
                try:
                    AxisTask._register_process(
                        process_file=scan.process_file,
                        entry=entry,
                        process=AxisTask,
                        results={"center_of_rotation": cor if cor is not None else "-"},
                        configuration=self._axis_params.to_dict(),
                        process_index=scan.pop_process_index(),
                        overwrite=True,
                    )
                except Exception as e:
                    logger.warning(f"Fail to register Axis process. Error is {e}")
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=AxisTask(
                    process_id=self.process_id,
                    inputs={
                        "data": None,
                        "serialize_output_data": False,
                    },
                ),
                state=DatasetState.SUCCEED,
            )
            self.scan_ready(scan=scan)

        elif self.isModeLocked():
            callback = functools.partial(self._updatePosition, scan)
            self._processingStack.add(
                data=scan, configuration=self._axis_params.to_dict(), callback=callback
            )
        else:
            self.activateWindow()
            self.raise_()
            self.show()

    def _updatePosition(self, scan):
        if scan.axis_params is not None:
            self._widget.setPosition(relative_value=scan.axis_params.relative_cor_value)

    @Inputs.data_recompute_axis
    def reprocess(self, scan):
        """Recompute the axis for scan"""
        if scan is not None:
            # for now The behavior for reprocessing is the sama as for processing
            if hasattr(scan, "instance"):
                self.process(scan.instance)
            else:
                self.process(scan)

    def close(self):
        self._processingStack.stop()
        self._processingStack.wait_computation_finished()
        self._widget = None
        self._processingStack = None
        super().close()

    def patch_calc_method(self, mode, function):
        self._patches.append((mode, function))

    def keyPressEvent(self, event):
        """The event has to be filtered since we have some children
        that can be edited using the 'enter' key as defining the cor manually
        (see #481)). As we are in a dialog this automatically trigger
        'accepted'. See https://forum.qt.io/topic/5080/preventing-enter-key-from-triggering-ok-in-qbuttonbox-in-particular-qlineedit-qbuttonbox/5
        """
        if event.key() != qt.Qt.Key_Enter:
            super().keyPressEvent(event)

    # expose API
    def setMode(self, mode):
        self._widget.setMode(mode=mode)

    def getMode(self):
        return self._widget.getMode()

    def getEstimatedCor(self):
        return self._widget.getEstimatedCor()

    def setEstimatedCor(self, value):
        self._widget.setEstimatedCor(value=value)

    def getAxisParams(self):
        return self._widget.getAxisParams()

    def setValueLock(self, lock: bool):
        self._widget.setValueLock(lock=lock)

    def _setModeLockFrmSettings(self, lock: bool):
        self._widget._setModeLockFrmSettings(lock=lock)

    def _setValueLockFrmSettings(self, lock: bool):
        self._widget._setValueLockFrmSettings(lock=lock)



from __future__ import annotations

import functools
import logging

from orangewidget import gui, settings
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.control.scantransfer
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from tomwer.core.process.control.scantransfer import ScanTransferTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datatransfert import DataTransfertSelector
from tomwer.utils import docstring

logger = logging.getLogger(__name__)


class DataTransfertOW(SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    name = "data transfer"
    id = "orange.widgets.tomwer.foldertransfert"
    description = "This widget insure data transfer of the received data "
    description += "to the given directory"
    icon = "icons/folder-transfert.svg"
    priority = 30
    keywords = [
        "tomography",
        "transfert",
        "cp",
        "copy",
        "move",
        "file",
        "tomwer",
        "folder",
    ]

    ewokstaskclass = tomwer.core.process.control.scantransfer.ScanTransferTask

    settingsHandler = CallbackSettingsHandler()

    want_main_area = True
    resizing_enabled = True

    dest_dir_settings = settings.Setting(str())
    """Parameters directly editabled from the TOFU interface"""

    scanready = qt.Signal(TomwerScanBase)
    """emit when scan ready"""

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        super().__init__(parent)

        self._destDir = None
        self._forceSync = False
        self._threads = []

        # define GUI
        self._widget = DataTransfertSelector(
            parent=self,
            rnice_option=True,
            default_root_folder=ScanTransferTask.getDefaultOutputDir(),
        )
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self._widget)

        # signal / SLOT connection
        self.settingsHandler.addCallback(self._updateSettingsVals)
        self._widget.sigSelectionChanged.connect(self._updateDestDir)

        # setting configuration
        if self.dest_dir_settings != "":
            self._widget.setFolder(self.dest_dir_settings)

    def _requestFolder(self):  # pragma: no cover
        """Launch a QFileDialog to ask the user the output directory"""
        dialog = qt.QFileDialog(self)
        dialog.setWindowTitle("Destination folder")
        dialog.setModal(1)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return None

        return dialog.selectedFiles()[0]

    def transfertDoneCallback(self, output_scan):
        if output_scan is None:
            return
        self.Outputs.data.send(output_scan)
        self.scanready.emit(output_scan)

    def _updateDestDir(self):
        self._destDir = self._widget.getFolder()

    def _updateSettingsVals(self):
        """function used to update the settings values"""
        self.dest_dir_settings = self._destDir

    @Inputs.data
    def process(self, scan):
        self._process(data=scan)

    def _process(self, data, move=False, force=True, noRsync=False):
        if data is None:
            return
        elif not isinstance(data, TomwerScanBase):
            raise TypeError("data is expected to be an instance of TomwerScanBase")

        inputs = {
            "data": data,
            "move": move,
            "overwrite": force,
            "noRsync": noRsync,
            "dest_dir": self._destDir,
            "block": self._forceSync,
            "serialize_output_data": False,
        }
        thread = ThreadDataTransfer(
            inputs=inputs,
            data=data,
            process=self,
        )
        try:
            process = ScanTransferTask(inputs=inputs)
        except Exception as e:
            logger.error(e)
        else:
            dest_dir = process.getDestinationDir(data.path, ask_for_output=False)
            if dest_dir is not None:
                thread.finished.connect(
                    functools.partial(
                        self.transfertDoneCallback,
                        data._deduce_transfert_scan(
                            process.getDestinationDir(data.path)
                        ),
                    )
                )
                thread.start()
            self._threads.append(thread)

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def setDestDir(self, dest_dir):
        self._destDir = dest_dir

    def setForceSync(self, sync):
        self._forceSync = sync

    def isCopying(self):
        # for now only move file is handled
        return False


class ThreadDataTransfer(qt.QThread):
    def __init__(self, data, inputs, process) -> None:
        super().__init__()
        self._inputs = inputs
        self._data = data
        self._process = process

    def run(self):
        try:
            process = ScanTransferTask(inputs=self._inputs)
            process.run()
        except Exception as e:
            logger.error(f"data transfer failed. Reason is {e}")
            ProcessManager().notify_dataset_state(
                dataset=self._data,
                process=self._process,
                state=DatasetState.FAILED,
                details=str(e),
            )
        else:
            ProcessManager().notify_dataset_state(
                dataset=self._data,
                process=self._process,
                state=DatasetState.SUCCEED,
            )



from __future__ import annotations

from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from tomwer.gui.configuration.level import ConfigurationLevel


class _FilteringObject:
    """
    Simple class to define if the widget should be visible or not if visible
    (at some moment the 'option' wudget should be visible) and if filtered.
    Should avoid some conflict when set set visible
    """

    def __init__(self, widget):
        self._widget = widget
        self._visible = True
        # is this
        self._filtered = False

    def setVisible(self, visible):
        self._visible = visible
        self._updateVisibility()

    def setFiltered(self, filtered):
        self._filtered = filtered
        self._updateVisibility()

    def _updateVisibility(self):
        self._widget.setVisible(self._visible and self._filtered)


class _NabuStageConfigBase:
    """Define interface for a specific nabu stage configuration widget
    (or a part of the stage configuration)
    """

    def __init__(self, stage: _NabuStages | str | None):
        if stage is None:
            self.__stage = None
        else:
            self.__stage = _NabuStages.from_value(stage)
        self._registeredWidgets = {}
        # list required widgets. Key is widget, value is the configuration
        # level

    def registerWidget(self, widget, config_level):
        """register a widget with a configuration level.

        :returns: _FilteringObject to use to define widget visibility
        """
        filteringObj = _FilteringObject(widget=widget)
        self._registeredWidgets[filteringObj] = ConfigurationLevel.from_value(
            config_level
        )
        return filteringObj

    def getConfiguration(self) -> dict:
        raise NotImplementedError("Base class")

    def setConfiguration(self, config) -> None:
        raise NotImplementedError("Base class")

    def getStage(self) -> _NabuStages | None:
        return self.__stage

    def setConfigurationLevel(self, level):
        for widget in self._registeredWidgets:
            filtered = self._registeredWidgets[widget] <= level
            widget.setFiltered(filtered)



# coding: utf-8

"""
This module is used to manage observations. Initially on files.
Observations are runned on a thread and run each n seconds.
They are manage by thread and signals
"""


import time

from silx.gui import qt


class QWaiterThread(qt.QThread):
    """simple thread wich wait for waitingTime to be finished"""

    def __init__(self, waitingTime):
        qt.QThread.__init__(self)
        self.waitingTime = waitingTime

    def run(self):
        time.sleep(self.waitingTime)



from __future__ import annotations

from silx.gui import qt

from nabu.stitching.config import StitchingType
from nabu.stitching import config as stitching_config
from nabu.stitching.config import (
    KEY_RESCALE_MAX_PERCENTILES,
    KEY_RESCALE_MIN_PERCENTILES,
    RESCALE_FRAMES,
    RESCALE_PARAMS,
    STITCHING_SECTION,
)
from nabu.stitching.utils import ShiftAlgorithm as StitchShiftAlgorithm

from nxtomomill.io.utils import convert_str_to_bool as _convert_str_to_bool

from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.gui.stitching.config.stitchingstrategies import StitchingStrategies
from tomwer.gui.stitching.z_stitching.fineestimation import _SliceGetter
from tomwer.gui.stitching.alignment import _AlignmentGroupBox
from tomwer.gui.stitching.config.axisparams import StitcherAxisParams
from tomwer.gui.stitching.normalization import NormalizationBySampleGroupBox
from tomwer.io.utils.utils import str_to_dict

from .utils import concatenate_dict


class StitchingOptionsWidget(qt.QWidget):
    """
    Widget to let the user define the different options for z-stitching such as which algorithm to search shift,
    which stitching strategy...
    """

    sigChanged = qt.Signal()
    """Signal emit when the options change"""
    sigSliceForPreviewChanged = qt.Signal(object)
    """Signal emit when the slice requested for the preview has changed. Parameter is a str or an int"""
    sigFlipLRChanged = qt.Signal(bool)
    """Signal emit when the request to flip LR frame has changed"""
    sigFlipUDChanged = qt.Signal(bool)
    """Signal emit when the request to flip UD frame has changed"""

    def __init__(
        self, first_axis: int, second_axis: int, parent=None, *args, **kwargs
    ) -> None:
        super().__init__(parent, *args, **kwargs)
        # stitching strategy (aka stitcher behavior)
        self.setLayout(qt.QFormLayout())
        self._stitchingStrategiesWidget = StitchingStrategies(
            parent=self, axis=first_axis
        )
        self._stitchingStrategiesWidget.setObjectName("strategy")
        self.layout().addRow(self._stitchingStrategiesWidget)
        # slice for preview
        self._previewSlices = _SliceGetter("middle", parent=self)
        self._previewSlices.setPlaceholderText(
            "slice index or one of ('middle', 'first', 'last')"
        )
        self._previewSlices.setToolTip(
            "expects a slice index (int > 0) or a str in ('first', 'middle', 'last')"
        )
        self.layout().addRow("slice for preview", self._previewSlices)

        # invert frame up - down
        self._flipLR_CB = qt.QCheckBox("flip frame left-right", self)
        self._flipLR_CB.setChecked(False)
        self._flipLR_CB.setToolTip(
            "Flip frame for stitching. This is mostly for volume has no metadata are existing to specify the direction of the frame"
        )
        self.layout().addRow(self._flipLR_CB)

        # invert frame left-right
        self._flipUD_CB = qt.QCheckBox("flip frame up-down", self)
        self._flipUD_CB.setChecked(False)
        self._flipUD_CB.setToolTip(
            "Flip frame for stitching. This is mostly for volume has no metadata are existing to specify the direction of the frame"
        )
        self.layout().addRow(self._flipUD_CB)

        # avoid data duplication
        self._avoidDataDuplication_CB = qt.QCheckBox("avoid data duplication", self)
        self._avoidDataDuplication_CB.setChecked(False)
        self._avoidDataDuplication_CB.setEnabled(False)
        self._avoidDataDuplication_CB.setToolTip(
            "Avoid data duplication (only enabled for stitching on reconstructed volumes)"
        )
        self.layout().addRow(self._avoidDataDuplication_CB)

        # alignment options
        self._alignmentGroup = _AlignmentGroupBox(self)
        self.layout().addRow(self._alignmentGroup)

        # slices to be reconstructed
        self._slices = _SlicesSelector(parent=self)
        self._slices.setToolTip(
            "for pre processing stitching those are projections and for post processing stitching those are volume slices"
        )
        self.layout().addRow(self._slices)

        # axis 0 params for shift search
        self._firstAxisGroup = qt.QGroupBox(f"axis {first_axis}", self)
        self._firstAxisGroup.setLayout(qt.QVBoxLayout())
        self.layout().addRow(self._firstAxisGroup)
        self._firstAxisShiftSearchParams = StitcherAxisParams(
            axis=first_axis, parent=self
        )
        self._firstAxisShiftSearchParams.layout().setContentsMargins(0, 0, 0, 0)
        self._firstAxisShiftSearchParams.layout().setSpacing(0)
        self._firstAxisGroup.layout().addWidget(self._firstAxisShiftSearchParams)

        # axis 2 params for shift search
        self._secondAxisGroup = qt.QGroupBox(f"axis {second_axis}", self)
        self._secondAxisGroup.setLayout(qt.QVBoxLayout())
        self.layout().addRow(self._secondAxisGroup)
        self._secondAxisShiftSearchParams = StitcherAxisParams(
            axis=second_axis, parent=self
        )
        self._secondAxisShiftSearchParams.layout().setContentsMargins(0, 0, 0, 0)
        self._secondAxisShiftSearchParams.layout().setSpacing(0)
        self._secondAxisGroup.layout().addWidget(self._secondAxisShiftSearchParams)
        # by default avoid doing shift along second axis by default
        self._secondAxisShiftSearchParams.setShiftSearchMethod(
            StitchShiftAlgorithm.NONE
        )

        # frame rescaling option
        self._rescalingWidget = RescalingWidget(parent=self)
        self.layout().addRow(self._rescalingWidget)

        # normalization by sample
        self._normalizationBySampleWidget = NormalizationBySampleGroupBox(parent=self)
        self._normalizationBySampleWidget.setChecked(False)
        self.layout().addRow(self._normalizationBySampleWidget)

        # connect signal / slot
        self._stitchingStrategiesWidget.sigChanged.connect(self._updated)
        self._previewSlices.editingFinished.connect(self._sliceForPreviewHasChanged)
        self._flipLR_CB.toggled.connect(self._flipLRHasChanged)
        self._flipUD_CB.toggled.connect(self._flipUDHasChanged)

    def _sliceForPreviewHasChanged(self):
        slice_for_preview = self.getSlicesForPreview()
        try:
            slice_for_preview = int(slice_for_preview)
        except ValueError:
            pass
        self.sigSliceForPreviewChanged.emit(slice_for_preview)
        self._updated()

    def _flipLRHasChanged(self):
        self.sigFlipLRChanged.emit(self._flipLR_CB.isChecked())

    def _flipUDHasChanged(self):
        self.sigFlipUDChanged.emit(self._flipUD_CB.isChecked())

    def _updated(self, *args, **kwargs):
        self.sigChanged.emit()

    def getSlicesForPreview(self):
        return self._previewSlices.text()

    def getSlices(self):
        slices = self._slices.getSlices()
        if slices == (0, -1, 1):
            return None
        else:
            return (str(slices[0]), str(slices[1]), str(slices[2]))

    def setSlices(self, slices: tuple):
        if isinstance(slices, str):
            slices = slices.replace(" ", "").split(":")
        if len(slices) > 2:
            step = int(slices[2])
        else:
            step = None
        self._slices.setSlices(int(slices[0]), int(slices[1]), step)

    def getAvoidDataDuplication(self) -> bool:
        return (
            self._avoidDataDuplication_CB.isChecked()
            and self._avoidDataDuplication_CB.isEnabled()
        )

    def setAvoidDataDuplication(self, avoid: bool):
        self._avoidDataDuplication_CB.setChecked(avoid)

    def getConfiguration(self) -> dict:
        slices = self.getSlices()
        if slices is None:
            slices = ""
        else:
            slices = ":".join(slices)
        res = {
            stitching_config.STITCHING_SECTION: {
                stitching_config.FLIP_LR: self._flipLR_CB.isChecked(),
                stitching_config.FLIP_UD: self._flipUD_CB.isChecked(),
                stitching_config.ALIGNMENT_AXIS_1_FIELD: self._alignmentGroup.getAlignmentAxis1().value,
                stitching_config.ALIGNMENT_AXIS_2_FIELD: self._alignmentGroup.getAlignmentAxis2().value,
                stitching_config.PAD_MODE_FIELD: self._alignmentGroup.getPadMode(),
                stitching_config.AVOID_DATA_DUPLICATION_FIELD: self.getAvoidDataDuplication(),
            },
            stitching_config.INPUTS_SECTION: {
                stitching_config.STITCHING_SLICES: slices,
            },
            stitching_config.NORMALIZATION_BY_SAMPLE_SECTION: self._normalizationBySampleWidget.getConfiguration(),
        }

        for ddict in (
            self._stitchingStrategiesWidget.getConfiguration(),
            self._firstAxisShiftSearchParams.getConfiguration(),
            self._secondAxisShiftSearchParams.getConfiguration(),
            self._rescalingWidget.getConfiguration(),
        ):
            res = concatenate_dict(res, ddict)
        return res

    def setConfiguration(self, config: dict):
        self._stitchingStrategiesWidget.setConfiguration(config)
        self._firstAxisShiftSearchParams.setConfiguration(config)
        self._secondAxisShiftSearchParams.setConfiguration(config)
        self._rescalingWidget.setConfiguration(config)
        slices = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.STITCHING_SLICES, ""
        )
        # slices
        if slices == "":
            slices = None
        if slices is not None:
            self.setSlices(slices)
        # flip_lr
        flip_lr = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.FLIP_LR, None
        )
        if flip_lr is not None:
            self._flipLR_CB.setChecked(flip_lr in (True, "True", 1, "1"))
        # flip_ud
        flip_ud = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.FLIP_UD, None
        )
        if flip_ud is not None:
            self._flipUD_CB.setChecked(flip_ud in (True, "True", 1, "1"))
        # avoid data duplication
        avoid_data_duplication = config.get(
            stitching_config.AVOID_DATA_DUPLICATION_FIELD, None
        )
        if avoid_data_duplication is not None:
            self.setAvoidDataDuplication(avoid=avoid_data_duplication)
        # alignment
        alignment_axis_1 = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.ALIGNMENT_AXIS_1_FIELD, None
        )
        if alignment_axis_1 is not None:
            self._alignmentGroup.setAlignmentAxis1(alignment_axis_1)
        alignment_axis_2 = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.ALIGNMENT_AXIS_2_FIELD, None
        )
        if alignment_axis_2 is not None:
            self._alignmentGroup.setAlignmentAxis2(alignment_axis_2)
        # pad_mode
        pad_mode = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.PAD_MODE_FIELD, None
        )
        if pad_mode is not None:
            self._alignmentGroup.setPadMode(pad_mode=pad_mode)

        # normalization by sample
        normalization_by_sample = config.get(
            stitching_config.NORMALIZATION_BY_SAMPLE_SECTION, None
        )
        if normalization_by_sample is not None:
            self._normalizationBySampleWidget.setConfiguration(normalization_by_sample)

    def _stitchingTypeChanged(self, stitching_type: str):
        stitching_type = StitchingType.from_value(stitching_type)
        self._alignmentGroup.setAlignmentAxis1Enabled(
            stitching_type is StitchingType.Z_POSTPROC
        )
        self._avoidDataDuplication_CB.setEnabled(
            stitching_type is StitchingType.Z_POSTPROC
        )

    def setConfigurationLevel(self, level: ConfigurationLevel):
        self._alignmentGroup.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._previewSlices.setVisible(level >= ConfigurationLevel.OPTIONAL)
        self._flipLR_CB.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._flipUD_CB.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._avoidDataDuplication_CB.setVisible(level >= ConfigurationLevel.OPTIONAL)
        self._rescalingWidget.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._firstAxisShiftSearchParams.setConfigurationLevel(level)
        self._secondAxisShiftSearchParams.setConfigurationLevel(level)
        self._normalizationBySampleWidget.setVisible(
            level >= ConfigurationLevel.ADVANCED
        )


class _SlicesSelector(qt.QGroupBox):
    """
    Widget to determine the slices values (to be stitched)
    """

    def __init__(self, parent=None) -> None:
        super().__init__("slices", parent)
        # start interface
        self.setLayout(qt.QHBoxLayout())
        self._startSliceCB = qt.QCheckBox("start", self)
        self.layout().addWidget(self._startSliceCB)
        self._startSliceSB = qt.QSpinBox(self)
        self._startSliceSB.setMinimum(0)
        self._startSliceSB.setMaximum(9999999)
        self._startSliceSB.setValue(0)
        self.layout().addWidget(self._startSliceSB)
        # stop interface
        self._stopSliceCB = qt.QCheckBox("stop", self)
        self.layout().addWidget(self._stopSliceCB)
        self._stopSliceSB = qt.QSpinBox(self)
        self._stopSliceSB.setMinimum(-1)
        self._stopSliceSB.setMaximum(9999999)
        self._stopSliceSB.setValue(-1)
        self.layout().addWidget(self._stopSliceSB)
        # step interface
        self._stepSliceLabel = qt.QLabel("step", self)
        self.layout().addWidget(self._stepSliceLabel)
        self._stepSliceSB = qt.QSpinBox(self)
        self._stepSliceSB.setMinimum(1)
        self._stepSliceSB.setMaximum(9999999)
        self._stepSliceSB.setValue(1)
        self.layout().addWidget(self._stepSliceSB)

        # connect signal / slot
        self._startSliceCB.toggled.connect(self._startSliceSB.setDisabled)
        self._stopSliceCB.toggled.connect(self._stopSliceSB.setDisabled)

        self._startSliceCB.setChecked(True)
        self._stopSliceCB.setChecked(True)

    def getSlices(self) -> tuple:
        if self._startSliceCB.isChecked():
            start = 0
        else:
            start = self._startSliceSB.value()
        if self._stopSliceCB.isChecked():
            stop = -1
        else:
            stop = self._stopSliceSB.value()
        step = self._stepSliceSB.value()
        return (start, stop, step)

    def setSlices(self, start: int, stop: int, step: int | None = None):
        start = int(start)
        stop = int(stop)
        if start == 0:
            self._startSliceCB.setChecked(True)
        else:
            self._startSliceCB.setChecked(False)
            self._startSliceSB.setValue(start)

        if stop == -1:
            self._stopSliceCB.setChecked(True)
        else:
            self._stopSliceCB.setChecked(False)
            self._stopSliceSB.setValue(stop)

        if step is not None:
            self._stepSliceSB.setValue(int(step))


class RescalingWidget(qt.QWidget):
    DEFAULT_MIN_PERCENTILE = 0

    DEFAULT_MAX_PERCENTILE = 100

    def __init__(self, parent, *args, **kwargs) -> None:
        super().__init__(parent, *args, **kwargs)
        self.setLayout(qt.QHBoxLayout())
        self._activatedCB = qt.QCheckBox("rescale frames", self)
        self.layout().addWidget(self._activatedCB)

        self._minPercentileQSB = qt.QSpinBox(self)
        self._minPercentileQSB.setRange(0, 100)
        self._minPercentileQSB.setPrefix("min:")
        self._minPercentileQSB.setSuffix("%")
        self._minPercentileQSB.setValue(self.DEFAULT_MIN_PERCENTILE)
        self.layout().addWidget(self._minPercentileQSB)

        self._maxPercentileQSB = qt.QSpinBox(self)
        self._maxPercentileQSB.setRange(0, 100)
        self._maxPercentileQSB.setPrefix("max:")
        self._maxPercentileQSB.setSuffix("%")
        self._maxPercentileQSB.setValue(self.DEFAULT_MAX_PERCENTILE)
        self.layout().addWidget(self._maxPercentileQSB)

        # set up
        self._activatedCB.setChecked(False)
        self._minPercentileQSB.setEnabled(False)
        self._maxPercentileQSB.setEnabled(False)

        # connect signal / slot
        self._activatedCB.toggled.connect(self._activationChanged)
        self._activatedCB.toggled.connect(self._activationChanged)

    def _activationChanged(self):
        self._minPercentileQSB.setEnabled(self._activatedCB.isChecked())
        self._maxPercentileQSB.setEnabled(self._activatedCB.isChecked())

    def getConfiguration(self):
        return {
            STITCHING_SECTION: {
                RESCALE_FRAMES: self._activatedCB.isChecked(),
                RESCALE_PARAMS: ";".join(
                    [
                        f"{KEY_RESCALE_MIN_PERCENTILES}={self._minPercentileQSB.value()}",
                        f"{KEY_RESCALE_MAX_PERCENTILES}={self._maxPercentileQSB.value()}",
                    ]
                ),
            }
        }

    def setConfiguration(self, config: dict):
        def cast_percentile(percentile) -> int:
            if isinstance(percentile, str):
                percentile.replace(" ", "").rstrip("%")
            return int(percentile)

        stitching_config = config.get(STITCHING_SECTION, {})
        rescale_params = str_to_dict(stitching_config.get(RESCALE_PARAMS, {}))
        rescale_min_percentile = rescale_params.get(KEY_RESCALE_MIN_PERCENTILES, None)
        if rescale_min_percentile is not None:
            rescale_min_percentile = cast_percentile(rescale_min_percentile)
            self._minPercentileQSB.setValue(rescale_min_percentile)
        rescale_max_percentile = rescale_params.get(KEY_RESCALE_MAX_PERCENTILES, None)
        if rescale_max_percentile is not None:
            rescale_max_percentile = cast_percentile(rescale_max_percentile)
            self._maxPercentileQSB.setValue(rescale_max_percentile)

        rescale = stitching_config.get(RESCALE_FRAMES, None)
        if rescale is not None:
            self._activatedCB.setChecked(_convert_str_to_bool(rescale))



from __future__ import annotations

import logging
import sys

import numpy
from silx.utils.enum import Enum as _Enum

_logger = logging.getLogger(__name__)


class ScoreMethod(_Enum):
    STD = "standard deviation"
    TV = "total variation"
    TV_INVERSE = "1 / (total variation)"
    STD_INVERSE = "1 / std"
    TOMO_CONSISTENCY = "tomo consistency"


class ComputedScore:
    def __init__(self, tv, std, tomo_consistency=None):
        self._tv = tv
        self._std = std
        self._tomo_consistency = tomo_consistency

    @property
    def total_variation(self):
        return self._tv

    @property
    def std(self):
        return self._std

    @property
    def tomo_consistency(self):
        return self._tomo_consistency

    def get(self, method: ScoreMethod):
        method = ScoreMethod.from_value(method)
        if method is ScoreMethod.TV:
            return self.total_variation / float(10e5)
        elif method is ScoreMethod.TV_INVERSE:
            return (1.0 / self.total_variation) * float(10e5)
        elif method is ScoreMethod.STD:
            return self.std
        elif method is ScoreMethod.STD_INVERSE:
            return 1.0 / self.std
        elif method is ScoreMethod.TOMO_CONSISTENCY:
            return self.tomo_consistency
        else:
            raise ValueError(f"{method} is an unrecognized method")

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, ComputedScore):
            return False
        else:
            return (
                self.total_variation == __value.total_variation
                and self.std == __value.std
                and self.tomo_consistency == __value.tomo_consistency
            )

    def __str__(self) -> str:
        return f"std: {self.std} - tv: {self.total_variation} - tomo-consistency: {self.tomo_consistency}"


def compute_tomo_consistency(
    data: numpy.ndarray,
    original_sinogram,
    angles,
    original_axis_position,
    detector_width,
):
    from silx.opencl.projection import Projection

    projector = Projection(
        slice_shape=data.shape,
        angles=angles,
        detector_width=detector_width,
        axis_position=original_axis_position,
    )
    sinogram = projector.projection(data)
    sinogram_n = (sinogram - sinogram.min()) / (sinogram.max() - sinogram.min())
    original_sinogram_n = (original_sinogram - original_sinogram.min()) / (
        original_sinogram.max() - original_sinogram.min()
    )
    return 1.0 / (numpy.sum(numpy.abs(sinogram_n - original_sinogram_n)) + 1)


def compute_score_contrast_std(data: numpy.ndarray):
    """
    Compute a contrast score by simply computing the standard deviation of
    the frame
    :param data: frame for which we should compute the score
    :return: score of the frame
    """
    if data is None:
        return None
    else:
        return data.std() * 100


def compute_tv_score(data: numpy.ndarray):
    """
    Compute the data score as image total variation

    :param data: frame for which we should compute the score
    :return: score of the frame
    """
    tv = numpy.sum(
        numpy.sqrt(
            numpy.gradient(data, axis=0) ** 2 + numpy.gradient(data, axis=1) ** 2
        )
    )
    return tv


_METHOD_TO_FCT = {
    ScoreMethod.STD: compute_score_contrast_std,
    ScoreMethod.TV: compute_tv_score,
}


def compute_score(
    data: numpy.ndarray,
    method: ScoreMethod,
    angles: list | None = None,
    original_sinogram: numpy.array | None = None,
    original_axis_position: float | None = None,
    detector_width: float | None = None,
) -> float | None:
    """

    :param data: frame for which we should compute the score
    :param method:
    :return: score of the frame
    """
    method = ScoreMethod.from_value(method)
    if data.ndim == 3:
        if data.shape[0] == 1:
            data = data.reshape(data.shape[1], data.shape[2])
        elif data.shape[2] == 1:
            data = data.reshape(data.shape[0], data.shape[1])
        else:
            raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")
    elif data.ndim == 2:
        pass
    else:
        raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")
    if method is ScoreMethod.TOMO_CONSISTENCY:
        try:
            return compute_tomo_consistency(
                data=data,
                angles=angles,
                original_sinogram=original_sinogram,
                original_axis_position=original_axis_position,
                detector_width=detector_width,
            )
        except ImportError:
            return None
        except Exception as e:
            _logger.warning(f"Fail to compute 'tomo consistency' score. Reason is {e}")
            return None
    else:
        fct = _METHOD_TO_FCT.get(method, None)

    if fct is not None:
        return fct(data)
    else:
        raise ValueError(f"{method} is not handled")


def get_disk_mask_radius(datasets_) -> int:
    """compute the radius to use for the mask"""
    radius = sys.maxsize
    # get min radius
    for _, (_, data) in datasets_.items():
        if data is None:
            continue
        assert data.ndim == 2, "data is expected to be 2D"
        min_ = numpy.array(data.shape).min()
        if radius >= min_:
            radius = min_
    return radius // 2


def apply_roi(data, radius, url) -> numpy.array:
    """compute the square included in the circle of radius and centered
    in the middle of the data"""
    half_width = int(radius / 2**0.5)
    center = numpy.array(data.shape[:]) // 2
    min_x, max_x = center[0] - half_width, center[0] + half_width
    min_y, max_y = center[1] - half_width, center[1] + half_width
    try:
        return data[min_y:max_y, min_x:max_x]
    except Exception:
        _logger.error(f"Fail to apply roi for {url.path()}. Take the entire dataset")
        return data



from __future__ import annotations

import numpy
from nxtomo.nxobject.nxdetector import ImageKey


def get_n_series(image_key_values: tuple | list, image_key_type: ImageKey) -> int:
    """
    Return the number of series of an image_key. Image key can be dark, flat, or projection.
    A series is defined as a contiguous elements in image_key_values

    :param image_key_values: list or tuple of image_keys to consider. Can be integers or tomoscan.esrf.scan.hdf5scan.ImageKey
    """
    image_key_type = ImageKey.from_value(image_key_type)
    if image_key_type is ImageKey.INVALID:
        raise ValueError(
            "we can't count Invalid image keys series because those are ignored from tomoscan"
        )
    image_key_values = [ImageKey.from_value(img_key) for img_key in image_key_values]

    # remove invalid frames
    image_key_values = numpy.array(
        image_key_values
    )  # for filtering invalid value a numpy aray is requested
    image_key_values = image_key_values[image_key_values != ImageKey.INVALID]

    n_series = 0
    is_in_a_serie = False
    for frame in image_key_values:
        if frame == image_key_type and not is_in_a_serie:
            is_in_a_serie = True
            n_series += 1
        elif frame != image_key_type:
            is_in_a_serie = False
    return n_series



from __future__ import annotations

from nabu.stitching.config import StitchingType
from silx.gui import qt
from silx.gui.utils import blockSignals

from tomoscan.scanbase import TomoScanBase

from tomwer.core.process.stitching.metadataholder import StitchingMetadata
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui.utils.unitsystem import MetricEntry, PixelEntry


class _AxisPositionGroup(qt.QGroupBox):
    sigUpdate = qt.Signal()
    """Signal emit when some data have been updated"""

    def __init__(self, axis: int, parent=None):
        assert axis in (0, 1, 2)
        super().__init__(title=f"axis {axis}", parent=parent)
        self.__axis = axis
        self._stitching_metadata = None

        self.setLayout(qt.QGridLayout())

        # pixel size
        self._pixelSizeRaw = MetricEntry(
            value="unknown", name="pixel size", parent=self, default_unit="mm"
        )
        self._pixelSizeRaw.setReadOnly(True)
        self.layout().addWidget(self._pixelSizeRaw, 0, 0, 1, 2)
        self._overridePixelSizeCB = qt.QCheckBox("overwrite", self)
        self.layout().addWidget(self._overridePixelSizeCB, 0, 2, 1, 1)
        self._pixelSizeOverride = MetricEntry(name="", parent=self, default_unit="mm")
        self.layout().addWidget(self._pixelSizeOverride, 0, 3, 1, 2)

        # raw position (in meter or millimeter)
        self._rawPositionMetric = MetricEntry(
            value="unknown",
            name=f"axis {axis} position (metric unit)",
            parent=self,
            default_unit="mm",
        )
        self._rawPositionMetric.setReadOnly(True)
        self.layout().addWidget(self._rawPositionMetric, 1, 0, 1, 2)
        self._overrideMetricPositionCB = qt.QCheckBox("overwrite", self)
        self.layout().addWidget(self._overrideMetricPositionCB, 1, 2, 1, 1)
        self._metricPositionOverride = MetricEntry(
            name="", parent=self, default_unit="mm"
        )
        self.layout().addWidget(self._metricPositionOverride, 1, 3, 1, 2)

        # raw position in pixel ...
        self._rawPxPosition = PixelEntry(name=f"axis {axis} position (pixel unit)")
        self._rawPxPosition.setReadOnly(True)
        self.layout().addWidget(self._rawPxPosition, 2, 0, 1, 2)
        self._overridePixelPositionCB = qt.QCheckBox("overwrite", self)
        self.layout().addWidget(self._overridePixelPositionCB, 2, 2, 1, 1)
        self._overridePxPosition = PixelEntry(
            name="",
            parent=self,
        )
        self.layout().addWidget(self._overridePxPosition, 2, 3, 1, 2)

        # utils to filter signals
        self._editingWidgets = (
            self._pixelSizeRaw,
            self._pixelSizeOverride,
            self._overridePixelPositionCB,
            self._overridePixelSizeCB,
            self._overrideMetricPositionCB,
            self._metricPositionOverride,
            self._rawPositionMetric,
            self._rawPxPosition,
            self._overridePxPosition,
        )

        # set up
        self._metricPositionOverride.setVisible(False)
        self._overridePxPosition.setVisible(False)
        self._pixelSizeOverride.setVisible(False)

        # connect signal / slot
        self._overrideMetricPositionCB.toggled.connect(
            self._metricPositionOverride.setVisible
        )
        self._overridePixelPositionCB.toggled.connect(
            self._overridePxPosition.setVisible
        )
        self._overridePixelSizeCB.toggled.connect(self._pixelSizeOverride.setVisible)

        self._overrideMetricPositionCB.toggled.connect(self._changed)
        self._overridePixelPositionCB.toggled.connect(self._changed)
        self._overridePixelSizeCB.toggled.connect(self._changed)
        self._metricPositionOverride.valueChanged.connect(self._changed)
        self._overridePxPosition.valueChanged.connect(self._changed)
        self._pixelSizeOverride.valueChanged.connect(self._changed)

    def _changed(self, *args, **kwargs):
        self._updateFinalPxPosition()
        self._updateTomoObjMetadata()

    def _updateTomoObjMetadata(self):
        # update the tomo obj metadata
        if self._stitching_metadata is not None:
            # disconnect the metadata signal to avoid sending several update message
            self._stitching_metadata.sigChanged.disconnect(self._loadMetadata)
            with blockSignals(self._stitching_metadata):
                self._stitching_metadata.setPixelOrVoxelSize(
                    value=self.getOverrridePixelSize(), axis=self.axis
                )
                self._stitching_metadata.setMetricPos(
                    self.getOverrrideMetricPosition(), axis=self.axis
                )
                self._stitching_metadata.setPxPos(
                    self.getOverrridePxPosition(), axis=self.axis
                )
            # make sure the signal of the tomo obj has been send (but this widget is not interested because
            # we did the update according to the widget.)
            self._stitching_metadata.sigChanged.emit()
            self._stitching_metadata.sigChanged.connect(self._loadMetadata)

    def _updateFinalPxPosition(self):
        # compute the final position
        if self._overridePixelPositionCB.isChecked():
            # if the pixel size is directly edited then ignore this one
            return None
        pixel_size_m = self.get_final_pixel_size()
        if pixel_size_m is None:
            return None

        position_m = self.get_final_metric_position()
        if position_m is None:
            return None

        position_px = int(position_m / pixel_size_m)
        self._rawPxPosition.setValue(position_px)

    @property
    def axis(self) -> int:
        return self.__axis

    def get_final_metric_position(self):
        """
        return the metric position to be used to compute the final pixel position
        """
        if self._overrideMetricPositionCB.isChecked():
            return self._metricPositionOverride.getValue()
        else:
            return self._overridePxPosition.getValue()

    def get_final_px_position(self):
        """
        return the pixel position to be used for stitching
        """
        if self._overridePixelPositionCB.isChecked():
            return self.getOverrridePxPosition()
        else:
            return self.getRawPxPosition()

    def get_final_pixel_size(self):
        if self._overridePixelPositionCB.isChecked():
            return self._pixelSizeOverride.getValue()
        else:
            return self._pixelSizeRaw.getValue()

    def setStitchingMetadata(self, stitching_metadata: QStitchingMetadata | None):
        """
        load used metadata from tomo_obj
        """
        if self._stitching_metadata is not None:
            self._stitching_metadata.sigChanged.disconnect(self._loadMetadata)
        # remove metadata because we are no more editing those
        self._stitching_metadata = None
        self.reset()
        self._stitching_metadata = stitching_metadata
        if self._stitching_metadata is not None:
            self._stitching_metadata.sigChanged.connect(self._loadMetadata)
            self._loadMetadata()

    def _loadMetadata(self):
        stitching_metadata = self._stitching_metadata
        tomo_obj = self._stitching_metadata.tomo_obj
        assert stitching_metadata is not None, "expects to have metadata"

        if isinstance(tomo_obj, TomoScanBase):
            # tODO: block all editor signal
            with blockSignals(*self._editingWidgets):
                # handle pixel size
                if self.axis == 0:
                    pixel_size = tomo_obj.y_pixel_size
                else:
                    pixel_size = tomo_obj.x_pixel_size
                if pixel_size is not None:
                    self.setRawPixelSize(pixel_size)
                override_pixel_size = stitching_metadata._pixel_or_voxel_size[self.axis]
                self._overridePixelSizeCB.setChecked(override_pixel_size is not None)
                if override_pixel_size is not None:
                    self.setOverrridePixelSize(override_pixel_size)
                # handle position in metric unit
                try:
                    raw_position_m = stitching_metadata.get_raw_position_m(
                        axis=self.axis
                    )
                except Exception:
                    raw_position_m = None
                if raw_position_m is not None:
                    self.setRawMetricPosition(position_m=raw_position_m)
                override_position_m = stitching_metadata._pos_as_m[self.axis]
                self._overrideMetricPositionCB.setChecked(
                    override_position_m is not None
                )
                if override_position_m is not None:
                    self.setOverrrideMetricPosition(position_m=override_position_m)

                # handle position in pixel
                l_pixel_size = override_pixel_size or pixel_size
                if raw_position_m is not None:
                    self.setRawPxPosition(int(raw_position_m / l_pixel_size))
                override_position_px = stitching_metadata._pos_as_px[self.axis]
                self._overridePixelPositionCB.setChecked(
                    override_position_px is not None
                )
                if override_position_px is not None:
                    self.setOverrridePxPosition(override_position_px)
        else:
            raise NotImplementedError(
                "loading volume metadata not handled for the moment"
            )

    def reset(self):
        self._rawPositionMetric.setValue("unknown")
        self._overrideMetricPositionCB.setChecked(False)
        self._rawPxPosition.setValue("unknown")
        self._overrideMetricPositionCB.setChecked(False)
        self._pixelSizeRaw.setValue("unknown")
        self._overridePixelSizeCB.setChecked(False)

    def getRawPxPosition(self):
        return self._rawPxPosition.getValue()

    def setRawPxPosition(self, position_px):
        self._rawPxPosition.setValue(position_px)

    def getOverrridePxPosition(self):
        if self._overridePixelPositionCB.isChecked():
            return self._overridePxPosition.getValue()
        else:
            return None

    def setOverrridePxPosition(self, position_px):
        if not self._overridePixelPositionCB.isChecked():
            with block_signals(self._overridePixelPositionCB):
                # avoid updating twice the stitching metadata value
                self._overridePixelPositionCB.setChecked(True)
        self._overridePxPosition.setValue(position_px)
        self._overridePxPosition.valueChanged.emit()

    def getOverrrideMetricPosition(self):
        if self._overrideMetricPositionCB.isChecked():
            return self._metricPositionOverride.getValue()
        else:
            return None

    def setOverrrideMetricPosition(self, position_m):
        if not self._overrideMetricPositionCB.isChecked():
            with block_signals(self._overrideMetricPositionCB):
                # avoid updating twice the stitching metadata value
                self._overrideMetricPositionCB.setChecked(True)
        self._metricPositionOverride.setValue(position_m)
        self._metricPositionOverride.valueChanged.emit()

    def getRawMetricPosition(self):
        return self._rawPositionMetric.getValue()

    def setRawMetricPosition(self, position_m, displayed_unit="mm"):
        self._rawPositionMetric.setValue(
            value_m=position_m, displayed_unit=displayed_unit
        )

    def getRawPixelSize(self):
        return self._pixelSizeRaw.getValue()

    def setRawPixelSize(self, pixel_size):
        self._pixelSizeRaw.setValue(pixel_size)

    def getOverrridePixelSize(self):
        if self._overridePixelSizeCB.isChecked():
            return self._pixelSizeOverride.getValue()
        else:
            return None

    def setOverrridePixelSize(self, pixel_size):
        if not self._overridePixelSizeCB.isChecked():
            with block_signals(self._overridePixelSizeCB):
                # avoid updating twice the stitching metadata value
                self._overridePixelSizeCB.setChecked(True)
        self._pixelSizeOverride.setValue(pixel_size)
        self._pixelSizeOverride.valueChanged.emit()


class TomoObjectPositionInfos(qt.QWidget):
    """
    Widget to define the TomoObject (scan or volume) position
    """

    def __init__(self, parent=None, stitching_metadata: StitchingMetadata = None):
        super().__init__(parent=parent)
        self._stitching_metadata = None
        self.setLayout(qt.QGridLayout())

        # axis 0 group
        self._axis_0_pos = _AxisPositionGroup(axis=0, parent=self)
        self.layout().addWidget(self._axis_0_pos, 1, 0, 1, 5)
        # axis 2 group
        self._axis_2_pos = _AxisPositionGroup(axis=2, parent=self)
        self.layout().addWidget(self._axis_2_pos, 2, 0, 1, 5)
        # spacer for style
        self._spacer = qt.QWidget(parent=self)
        self._spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacer, 99, 99, 1, 1)

        # set up
        self.setStitchingMetadata(stitching_metadata)

    def getStitchingMetadata(self):
        return self._stitching_metadata

    def reset(self):
        self._axis_0_pos.reset()
        self._axis_2_pos.reset()

    def clean(self):
        self.reset()
        self._stitching_metadata = None

    def setStitchingMetadata(self, stitching_metadata: StitchingMetadata | None = None):
        if stitching_metadata is not None and not isinstance(
            stitching_metadata, StitchingMetadata
        ):
            raise TypeError(
                "stitching_metadata is expected to be an instance of {StitchingMetadata}"
            )
        else:
            self._stitching_metadata = stitching_metadata
            self._loadMetadata(stitching_metadata)

    def _loadMetadata(self, metadata: StitchingMetadata):
        self._axis_0_pos.setStitchingMetadata(metadata)
        self._axis_2_pos.setStitchingMetadata(metadata)

    def updateStitchingType(self, stitching_type: StitchingType):
        stitching_type = StitchingType.from_value(stitching_type)
        if stitching_type is StitchingType.Z_POSTPROC:
            pixel_or_voxel_label = "voxel size"
        elif stitching_type is StitchingType.Z_PREPROC:
            pixel_or_voxel_label = "pixel size"
        else:
            raise NotImplementedError
        self._axis_0_pos._pixelSizeRaw.setLabelText(pixel_or_voxel_label)



from __future__ import annotations

import fileinput
import logging
import os

import fabio.edfimage
from lxml import etree
from pyunitsystem.metricsystem import MetricSystem


_logger = logging.getLogger(__name__)


def _getInformation(scan, refFile, information, _type, aliases=None):
    """
    Parse files contained in the given directory to get the requested
    information

    :param scan: directory containing the acquisition. Must be an absolute path
    :param refFile: the refXXXX_YYYY which should contain information about the
                    scan.
    :return: the requested information or None if not found
    """

    def parseRefFile(filePath):
        header = fabio.open(filePath).header
        for k in aliases:
            if k in header:
                return _type(header[k])
        return None

    def parseXMLFile(filePath):
        try:
            for alias in info_aliases:
                tree = etree.parse(filePath)
                elmt = tree.find("acquisition/" + alias)
                if elmt is None:
                    continue
                else:
                    info = _type(elmt.text)
                    if info == -1:
                        return None
                    else:
                        return info
        except etree.XMLSyntaxError as e:
            _logger.warning(e)
            return None

    def parseInfoFile(filePath):
        def extractInformation(text, alias):
            text = text.replace(alias, "")
            text = text.replace("\n", "")
            text = text.replace(" ", "")
            text = text.replace("=", "")
            return _type(text)

        info = None
        f = open(filePath, "r")
        line = f.readline()
        while line:
            for alias in info_aliases:
                if alias in line:
                    info = extractInformation(line, alias)
                    break
            line = f.readline()
        f.close()
        return info

    info_aliases = [information]
    if aliases is not None:
        assert type(aliases) in (tuple, list)
        [info_aliases.append(alias) for alias in aliases]

    if not os.path.isdir(scan):
        return None

    if refFile is not None and os.path.isfile(refFile):
        try:
            info = parseRefFile(refFile)
        except IOError as e:
            _logger.warning(e)
        else:
            if info is not None:
                return info

    baseName = os.path.basename(scan)
    infoFiles = [os.path.join(scan, baseName + ".info")]
    infoOnDataVisitor = infoFiles[0].replace("lbsram", "", 1)
    # hack to check in lbsram, would need to be removed to add some consistency
    if os.path.isfile(infoOnDataVisitor):
        infoFiles.append(infoOnDataVisitor)
    for infoFile in infoFiles:
        if os.path.isfile(infoFile) is True:
            info = parseInfoFile(infoFile)
            if info is not None:
                return info

    xmlFiles = [os.path.join(scan, baseName + ".xml")]
    xmlOnDataVisitor = xmlFiles[0].replace("lbsram", "", 1)
    # hack to check in lbsram, would need to be removed to add some consistency
    if os.path.isfile(xmlOnDataVisitor):
        xmlFiles.append(xmlOnDataVisitor)
    for xmlFile in xmlFiles:
        if os.path.isfile(xmlFile) is True:
            info = parseXMLFile(xmlFile)
            if info is not None:
                return info

    return None


def getClosestEnergy(scan, refFile=None):
    """
    Parse files contained in the given directory to get information about the
    incoming energy for the serie `iSerie`

    :param scan: directory containing the acquisition
    :param refFile: the refXXXX_YYYY which should contain information about the
                    energy.
    :return: the energy in keV or none if no energy found
    """
    return _getInformation(
        os.path.abspath(scan),
        refFile,
        information="Energy",
        aliases=["energy", "ENERGY"],
        _type=float,
    )


def getTomo_N(scan):
    """Return the number of radio taken"""
    return _getInformation(
        os.path.abspath(scan),
        refFile=None,
        information="TOMO_N",
        _type=int,
        aliases=["tomo_N", "Tomo_N"],
    )


def getDARK_N(scan):
    return _getInformation(
        os.path.abspath(scan),
        refFile=None,
        information="DARK_N",
        _type=int,
        aliases=["dark_N"],
    )


def rebaseParFile(_file, oldfolder, newfolder):
    """Update the given .par file to replace oldfolder location by the newfolder.

    .. warning:: make the replacement in place.

    :param _file: par file to update
    :param oldfolder: previous location of the .par file
    :param newfolder: new location of the .par file
    """
    with fileinput.FileInput(_file, inplace=True, backup=".bak") as parfile:
        for line in parfile:
            line = line.rstrip().replace(oldfolder, newfolder, 1)
            print(line)


def getDim1Dim2(scan: str) -> tuple[int]:
    """

    :param scan: path to the acquisition
    :return: detector definition
    """
    d1 = _getInformation(
        scan=scan,
        refFile=None,
        information="Dim_1",
        aliases=["projectionSize/DIM_1"],
        _type=int,
    )
    d2 = _getInformation(
        scan=scan,
        refFile=None,
        information="Dim_2",
        aliases=["projectionSize/DIM_2"],
        _type=int,
    )
    return d1, d2


def getParametersFromParOrInfo(_file):
    """
    Create a dictionary from the file with the information name as keys and
    their values as values
    """
    assert os.path.exists(_file) and os.path.isfile(_file)
    ddict = {}
    f = open(_file, "r")
    lines = f.readlines()
    for line in lines:
        if "=" not in line:
            continue
        line_str = line.replace(" ", "")
        line_str = line_str.rstrip("\n")
        # remove on the line comments
        if "#" in line_str:
            line_str = line_str.split("#")[0]
        if line_str == "":
            continue
        try:
            key, value = line_str.split("=")
        except ValueError:
            _logger.error('fail to extract information from "%s"' % line_str)
        else:
            ddict[key.lower()] = value
    return ddict


def getFirstProjFile(scan):
    """Return the first .edf containing a projection"""
    if os.path.isdir(scan) is False:
        return None
    files = sorted(os.listdir(scan))

    while (
        len(files) > 0
        and (files[0].startswith(os.path.basename(scan)) and files[0].endswith(".edf"))
        is False
    ):
        files.remove(files[0])

    if len(files) > 0:
        return os.path.join(scan, files[0])
    else:
        return None


def getPixelSize(scan) -> float | None:
    """
    Try to retrieve the pixel size from the set of files.

    :return: the pixel size in meter or None
    """
    if os.path.isdir(scan) is False:
        return None
    value = _getInformation(
        scan=scan,
        refFile=None,
        information="PixelSize",
        _type=float,
        aliases=["pixelSize"],
    )
    if value is None:
        parFile = os.path.join(scan, os.path.basename(scan) + ".par")
        if os.path.exists(parFile):
            ddict = getParametersFromParOrInfo(parFile)
            if "IMAGE_PIXEL_SIZE_1".lower() in ddict:
                value = float(ddict["IMAGE_PIXEL_SIZE_1".lower()])
    # for now pixel size are stored in microns. We want to return them in meter
    if value is not None:
        return value * MetricSystem.MICROMETER.value
    else:
        return None



# coding: utf-8
from __future__ import annotations


import logging
import time

from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.utils.deprecation import deprecated_warning, deprecated

_logger = logging.getLogger(__name__)


class TimerTask(
    Task,
    input_names=("data", "wait"),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Simple timer / time out - function
    """

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if inputs is None:
            inputs = {}

    @property
    @deprecated(replacement="task.wait", since_version="1.2")
    def waiting_time(self):
        return self.inputs.wait

    @waiting_time.setter
    @deprecated(replacement="task.wait", since_version="1.2")
    def waiting_time(self, wait):
        self.inputs.wait = wait

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be a dict or an instance of TomwerScanBase. Not {type(scan)}"
            )
        time.sleep(self.inputs.wait)
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan


class Timer(TimerTask):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.control.timer.Timer",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="TimerTask",
        )
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)
